#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
examples.halving-calculator.py: Demonstrate use of the MMGen asyncio/aiohttp JSON-RPC interface
"""

import time

from mmgen.cfg import Config
from mmgen.util import async_run

opts_data = {
	'text': {
		'desc': 'Estimate date of next block subsidy halving',
		'usage':'[opts]',
		'options': """
-h, --help          Print this help message
--, --longhelp      Print help message for long (global) options
-s, --sample-size=N Specify sample block range for block discovery time
                    estimate
""",
	'notes': """
Requires a running coin daemon
Specify coin with --coin=btc (default)/--coin=bch/--coin=ltc
If necessary, invoke with --rpc-host/--rpc-port/--rpc-user/--rpc-password
Specify aiohttp backend with --rpc-backend=aiohttp (Linux only)

A more full-featured version of this program can be found in the
mmgen-node-tools repository.
"""
	}
}

cfg = Config(opts_data=opts_data)

def date(t):
	return '{}-{:02}-{:02} {:02}:{:02}:{:02}'.format(*time.gmtime(t)[:6])

def dhms(t):
	t, neg = (-t, '-') if t < 0 else (t, ' ')
	return f'{neg}{t//60//60//24} days, {t//60//60%24:02}:{t//60%60:02}:{t%60:02} h/m/s'

def time_diff_warning(t_diff):
	if abs(t_diff) > 60*60:
		print('Warning: block tip time is {} {} clock time!'.format(
			dhms(abs(t_diff)),
			('behind', 'ahead of')[t_diff<0]))

async def main():

	proto = cfg._proto

	from mmgen.rpc import rpc_init
	c = await rpc_init(cfg, proto, ignore_wallet=True)

	tip = await c.call('getblockcount')
	assert tip > 1, 'block tip must be > 1'
	remaining = proto.halving_interval - tip % proto.halving_interval
	sample_size = int(cfg.sample_size) if cfg.sample_size else min(tip-1, max(remaining, 144))

	# aiohttp backend will perform these two calls concurrently:
	cur, old = await c.gathered_call('getblockstats', ((tip,), (tip - sample_size,)))

	clock_time = int(time.time())
	time_diff_warning(clock_time - cur['time'])

	bdr = (cur['time'] - old['time']) / sample_size
	t_rem = remaining * int(bdr)
	sub = proto.coin_amt(cur['subsidy'], from_unit='satoshi' if isinstance(cur['subsidy'], int) else None)

	print(
		f'Current block:      {tip}\n'
		f'Next halving block: {tip + remaining}\n'
		f'Blocks until halving: {remaining}\n'
		f'Current block subsidy: {str(sub).rstrip("0")} {proto.coin}\n'
		f'Current block discovery rate (over last {sample_size} blocks): {bdr/60:0.1f} minutes\n'
		f'Current clock time (UTC): {date(clock_time)}\n'
		f'Est. halving date (UTC):  {date(cur["time"] + t_rem)}\n'
		f'Est. time until halving: {dhms(cur["time"] + t_rem - clock_time)}\n'
	)

async_run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
subseed: Subseed classes and methods for the MMGen suite
"""

from .color import green
from .util import msg_r, msg, die, make_chksum_8
from .objmethods import MMGenObject, HiliteStr, InitErrors
from .obj import MMGenRange, IndexedDict, ImmutableAttr
from .seed import SeedBase, SeedID

class SubSeedIdxRange(MMGenRange):
	min_idx = 1
	max_idx = 1000000

class SubSeedIdx(HiliteStr, InitErrors):
	color = 'red'
	trunc_ok = False
	def __new__(cls, s):
		if isinstance(s, cls):
			return s
		try:
			assert isinstance(s, str), 'not a string or string subclass'
			idx = s[:-1] if s[-1] in 'SsLl' else s
			from .util import is_int
			assert is_int(idx), "valid format: an integer, plus optional letter 'S', 's', 'L' or 'l'"
			idx = int(idx)
			assert idx >= SubSeedIdxRange.min_idx, f'subseed index < {SubSeedIdxRange.min_idx:,}'
			assert idx <= SubSeedIdxRange.max_idx, f'subseed index > {SubSeedIdxRange.max_idx:,}'

			sstype, ltr = ('short', 'S') if s[-1] in 'Ss' else ('long', 'L')
			me = str.__new__(cls, str(idx)+ltr)
			me.idx = idx
			me.type = sstype
			return me
		except Exception as e:
			return cls.init_fail(e, s)

class SubSeed(SeedBase):

	idx    = ImmutableAttr(int, typeconv=False)
	nonce  = ImmutableAttr(int, typeconv=False)
	ss_idx = ImmutableAttr(SubSeedIdx)
	max_nonce = 1000

	def __init__(self, parent_list, idx, nonce, length):
		self.idx = idx
		self.nonce = nonce
		self.ss_idx = str(idx) + {'long': 'L', 'short': 'S'}[length]
		self.parent_list = parent_list
		SeedBase.__init__(
			self,
			parent_list.parent_seed.cfg,
			seed_bin=self.make_subseed_bin(parent_list, idx, nonce, length))

	@staticmethod
	def make_subseed_bin(parent_list, idx:int, nonce:int, length:str):
		seed = parent_list.parent_seed
		short = {'short': True, 'long': False}[length]
		# field maximums: idx: 4294967295 (1000000), nonce: 65535 (1000), short: 255 (1)
		scramble_key  = idx.to_bytes(4, 'big') + nonce.to_bytes(2, 'big') + short.to_bytes(1, 'big')
		from .crypto import Crypto
		return Crypto(parent_list.parent_seed.cfg).scramble_seed(
			seed.data, scramble_key)[:16 if short else seed.byte_len]

class SubSeedList(MMGenObject):
	have_short = True
	nonce_start = 0
	debug_last_share_sid_len = 3
	dfl_len = 100

	def __init__(self, parent_seed, *, length=None):
		self.member_type = SubSeed
		self.parent_seed = parent_seed
		self.data = {'long': IndexedDict(), 'short': IndexedDict()}
		self.len = length or self.dfl_len

	def __len__(self):
		return len(self.data['long'])

	def get_subseed_by_ss_idx(self, ss_idx_in, *, print_msg=False):
		ss_idx = SubSeedIdx(ss_idx_in)
		if print_msg:
			msg_r('{} {} of {}...'.format(
				green('Generating subseed'),
				ss_idx.hl(),
				self.parent_seed.sid.hl(),
			))

		if ss_idx.idx > len(self):
			self._generate(ss_idx.idx)

		sid = self.data[ss_idx.type].key(ss_idx.idx-1)
		idx, nonce = self.data[ss_idx.type][sid]
		if idx != ss_idx.idx:
			die(3, "{} != {}: self.data[{t!r}].key(i) does not match self.data[{t!r}][i]!".format(
				idx,
				ss_idx.idx,
				t = ss_idx.type))

		if print_msg:
			msg(f'\b\b\b => {SeedID.hlc(sid)}')

		seed = self.member_type(self, idx, nonce, length=ss_idx.type)
		assert seed.sid == sid, f'{seed.sid} != {sid}: Seed ID mismatch!'
		return seed

	def get_subseed_by_seed_id(self, sid, *, last_idx=None, print_msg=False):

		def get_existing_subseed_by_seed_id(sid):
			for k in ('long', 'short') if self.have_short else ('long',):
				if sid in self.data[k]:
					idx, nonce = self.data[k][sid]
					return self.member_type(self, idx, nonce, length=k)

		def do_msg(subseed):
			if print_msg:
				self.parent_seed.cfg._util.qmsg('{} {} ({}:{})'.format(
					green('Found subseed'),
					subseed.sid.hl(),
					self.parent_seed.sid.hl(),
					subseed.ss_idx.hl(),
				))

		if last_idx is None:
			last_idx = self.len

		subseed = get_existing_subseed_by_seed_id(sid)
		if subseed:
			do_msg(subseed)
			return subseed

		if len(self) >= last_idx:
			return None

		self._generate(last_idx, last_sid=sid)

		subseed = get_existing_subseed_by_seed_id(sid)
		if subseed:
			do_msg(subseed)
			return subseed

	def _collision_debug_msg(self, sid, idx, nonce, *, nonce_desc='nonce', debug_last_share=False):
		slen = 'short' if sid in self.data['short'] else 'long'
		m1 = f'add_subseed(idx={idx},{slen}):'
		if sid == self.parent_seed.sid:
			m2 = f'collision with parent Seed ID {sid},'
		else:
			if debug_last_share:
				sl = self.debug_last_share_sid_len
				colliding_idx = [d[:sl] for d in self.data[slen].keys].index(sid[:sl]) + 1
				sid = sid[:sl]
			else:
				colliding_idx = self.data[slen][sid][0]
			m2 = f'collision with ID {sid} (idx={colliding_idx},{slen}),'
		msg(f'{m1:30} {m2:46} incrementing {nonce_desc} to {nonce+1}')

	def _generate(self, last_idx=None, *, last_sid=None):

		if last_idx is None:
			last_idx = self.len

		first_idx = len(self) + 1

		if first_idx > last_idx:
			return None

		if last_sid is not None:
			last_sid = SeedID(sid=last_sid)

		def add_subseed(idx, length):
			for nonce in range(self.nonce_start, self.member_type.max_nonce+1): # handle SeedID collisions
				sid = make_chksum_8(self.member_type.make_subseed_bin(self, idx, nonce, length))
				if sid in self.data['long'] or sid in self.data['short'] or sid == self.parent_seed.sid:
					if self.parent_seed.cfg.debug_subseed: # should get ≈450 collisions for first 1,000,000 subseeds
						self._collision_debug_msg(sid, idx, nonce)
				else:
					self.data[length][sid] = (idx, nonce)
					return last_sid == sid
			# must exit here, as this could leave self.data in inconsistent state
			die('SubSeedNonceRangeExceeded', 'add_subseed(): nonce range exceeded')

		for idx in SubSeedIdxRange(first_idx, last_idx).iterate():
			match1 = add_subseed(idx, 'long')
			match2 = add_subseed(idx, 'short') if self.have_short else False
			if match1 or match2:
				break

	def format(self, first_idx, last_idx):

		r = SubSeedIdxRange(first_idx, last_idx)

		if len(self) < last_idx:
			self._generate(last_idx)

		fs1 = '{:>18} {:>18}\n'
		fs2 = '{i:>7}L: {:8} {i:>7}S: {:8}\n'

		hdr = f'    Parent Seed: {self.parent_seed.sid.hl()} ({self.parent_seed.bitlen} bits)\n\n'
		hdr += fs1.format('Long Subseeds', 'Short Subseeds')
		hdr += fs1.format('-------------', '--------------')

		sl = self.data['long'].keys
		ss = self.data['short'].keys
		body = (fs2.format(sl[n-1], ss[n-1], i=n) for n in r.iterate())

		return hdr + ''.join(body)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.ethdev: Ethdev tests for the cmdtest.py test suite
"""

import sys, time, os, re, shutil, asyncio, json
from decimal import Decimal
from collections import namedtuple
from subprocess import run, PIPE, DEVNULL
from pathlib import Path

from mmgen.color import red, yellow, blue, cyan, orange, set_vt100
from mmgen.util import msg, msg_r, rmsg, die
from mmgen.proto.eth.util import compute_contract_addr

from ..include.common import (
	cfg,
	check_solc_ver,
	omsg,
	omsg_r,
	imsg,
	imsg_r,
	joinpath,
	read_from_file,
	write_to_file,
	cmp_or_die,
	strip_ansi_escapes,
	silence,
	end_silence,
	gr_uc,
	stop_test_daemons)

from .include.common import (
	ref_dir,
	dfl_words_file,
	dfl_sid,
	tx_comment_jp,
	tx_comment_lat_cyr_gr,
	tw_comment_zh,
	tw_comment_lat_cyr_gr,
	get_file_with_ext,
	ok_msg,
	Ctrl_U,
	cleanup_env,
	thorchain_router_addr_file)
from .include.proxy import TestProxy

from .base import CmdTestBase
from .shared import CmdTestShared
from .httpd.etherscan import EtherscanServer

etherscan_server = EtherscanServer()

del_addrs = ('4', '1')

# The OpenEthereum dev address with lots of coins.  Create with "ethkey -b info ''":
dfl_devaddr = '00a329c0648769a73afac7f9381e08fb43dbea72'
dfl_devkey = '4d5db4107d237df6a3d58ee5f70ae63d73d7658d4026f2eefd2f204c81682cb7'
dfl_devkey_fn = 'dfl.devkey'

def get_reth_dev_keypair(cfg):
	from mmgen.bip39 import bip39
	from mmgen.bip_hd import MasterNode
	mn = 'test test test test test test test test test test test junk' # See ‘reth node --help’
	seed = bip39().generate_seed(mn.split())
	m = MasterNode(cfg, seed)
	node = m.to_chain(idx=0, coin='eth').derive_private(0)
	return (node.key.hex(), node.address)

reth_dev_amt = 1_000_000
reth_devkey_fn = 'reth.devkey'

burn_addr  = 'deadbeef'*5
burn_addr2 = 'beadcafe'*5

amt1 = '999777.12345689012345678'
amt2 = '888.111122223333444455'

def set_vbals(daemon_id):
	global vbal1, vbal2, vbal3, vbal4, vbal5, vbal6, vbal7, vbal9
	if daemon_id == 'geth':
		vbal1 = '1.2288334'
		vbal2 = '99.996560752'
		vbal3 = '1.2314176'
		vbal4 = '127.0287834'
		vbal5 = '999904.14775104212345678'
		vbal6 = '999904.14880104212345678'
		vbal7 = '999902.91891764212345678'
		vbal9 = '1.2262504'
	elif daemon_id == 'reth':
		vbal1 = '1.2288334'
		vbal2 = '99.996560752'
		vbal3 = '1.23142525'
		vbal3 = '1.2314176'
		vbal4 = '127.0287834'
		vbal5 = '999904.14775104212345678'
		vbal6 = '999904.14880104212345678'
		vbal7 = '999902.91891764212345678'
		vbal9 = '1.2262504'
	else:
		vbal1 = '1.2288396'
		vbal2 = '99.997088092'
		vbal3 = '1.23142525'
		vbal3 = '1.2314246'
		vbal4 = '127.0287896'
		vbal5 = '999904.14828458212345678'
		vbal6 = '999904.14933458212345678'
		vbal7 = '999902.91944498212345678'
		vbal9 = '1.226261'

coin = cfg.coin

class CmdTestEthdevMethods:

	def _del_addr(self, addr):
		t = self.spawn('mmgen-tool', self.eth_opts + ['remove_address', addr])
		t.expect(f"'{addr}' deleted")
		return t

	def _addrgen(self, addrs='1-3,11-13,21-23', no_msg=False):
		t = self.spawn(
			'mmgen-addrgen',
			[f'--coin={self.proto.coin}'] + self.eth_opts + [dfl_words_file, addrs],
			no_msg = no_msg,
			no_passthru_opts = True)
		t.written_to_file('Addresses')
		return t

	def _addrimport(
			self,
			ext       = '21-23]{}.regtest.addrs',
			expect    = '9/9',
			add_args  = [],
			bad_input = False,
			exit_val  = None):
		ext = ext.format('-α' if self.cfg.debug_utf8 else '')
		fn = self.get_file_with_ext(ext, no_dot=True, delete=False)
		t = self.spawn('mmgen-addrimport', ['--regtest=1'] + add_args + [fn], exit_val=exit_val)
		if bad_input:
			return t
		t.expect('Importing')
		t.expect(expect)
		return t

	def _create_tx(self, *, fee, args, add_opts=[]):
		return self.txcreate_ui_common(
			self.spawn('mmgen-txcreate', add_opts + ['-B'] + args),
			caller            = 'txcreate',
			input_sels_prompt = 'to spend from',
			inputs            = '1',
			file_desc         = 'transaction',
			interactive_fee   = fee,
			fee_desc          = 'transaction fee or gas price')

	def _send_tx(self, *, desc='transaction', add_opts=[]):
		t = self.spawn('mmgen-txsend', add_opts, no_passthru_opts=['coin'])
		t.view_tx('t')
		t.expect('(y/N): ', 'n')
		self._do_confirm_send(t, quiet=True)
		t.written_to_file(f'Sent {desc}')
		return t

	def _txdo(self, *, args, acct, fee='50G'):
		t = self.txcreate(
			args,
			acct            = acct,
			caller          = 'txdo',
			no_read         = True,
			bad_input_sels  = False,
			interactive_fee = fee,
			print_listing   = False)
		t.written_to_file('Signed transaction')
		t.expect('confirm: ', 'YES\n')
		return t

	def _txbump(self, *, fee, ext, add_opts=[], add_args=[]):
		ext = ext.format('-α' if self.cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		t = self.spawn('mmgen-txbump', self.eth_opts + add_opts + ['--yes', txfile] + add_args)
		t.expect('or gas price: ', fee+'\n')
		return t

	def _tx_receipt(self, ext='{}.regtest.sigtx'):
		self.mining_delay()
		return self.txsend(
			ext = ext,
			add_args = ['--receipt'],
			return_early = True,
			env = cleanup_env(cfg=self.cfg))

	def fund_mmgen_addr1(self):
		return self._fund_mmgen_addr(arg=f'{dfl_sid}:E:1,{self.fund_amt}')

	def fund_mmgen_addr2(self):
		return self._fund_mmgen_addr(arg=f'{dfl_sid}:E:11,{self.fund_amt}')

	def fund_mmgen_addr3(self):
		return self._fund_mmgen_addr(arg=f'{dfl_sid}:E:21,{self.fund_amt}')

	def _fund_mmgen_addr(self, arg):
		self.mining_delay()
		return self._txdo(
			args = [f'--keys-from-file={joinpath(self.tmpdir, dfl_devkey_fn)}', arg, dfl_words_file],
			acct = '10')

	def _bal_check(self, *, pat, add_opts=[]):
		self.mining_delay()
		t = self.spawn('mmgen-tool', ['--regtest=1'] + add_opts + ['twview', 'wide=1'])
		text = t.read(strip_color=True)
		assert re.search(pat, text, re.DOTALL), f'output failed to match regex {pat}'
		return t

	def _token_addrgen(self, *, mm_idxs, naddrs):
		self.spawn(msg_only=True)
		for idx in mm_idxs:
			t = self._addrgen(addrs=f'{idx}-{idx+naddrs-1}', no_msg=True)
		return t

	def _token_addrimport(self, addr_file, addr_range, expect, extra_args=[]):
		token_addr = self.read_from_tmpfile(addr_file).strip()
		return self._addrimport(
			ext      = f'[{addr_range}]{{}}.regtest.addrs',
			expect   = expect,
			add_args = ['--token-addr='+token_addr]+extra_args)

	def _get_contract_address(self, deployer_addr):
		t = self.spawn(
			'mmgen-cli',
			['--regtest=1', 'eth_getTransactionCount', '0x'+deployer_addr, 'pending'],
			env = cleanup_env(cfg=self.cfg),
			silent = True,
			no_msg = True)
		nonce = t.read().strip()
		imsg(f'Nonce: {red(nonce)}')
		ret = compute_contract_addr(self.cfg, deployer_addr, int(nonce, 16))
		imsg(f'Computed contract address: {blue(ret)}')
		return ret

	async def _token_deploy(
			self,
			*,
			key,
			gas,
			mmgen_cmd = 'txdo',
			gas_price = '8G',
			fn = None,
			num = None):

		keyfile = joinpath(self.tmpdir, dfl_devkey_fn)
		fn = fn or joinpath(self.tmpdir, 'mm'+str(num), key+'.bin')
		args = [
			'-B',
			f'--fee={gas_price}',
			f'--gas={gas}',
			f'--contract-data={fn}',
			f'--inputs={dfl_devaddr}',
			'--yes'
		] + (['--wait'] if mmgen_cmd == 'txdo' else [])

		contract_addr = self._get_contract_address(dfl_devaddr)
		if key == 'Token':
			self.write_to_tmpfile(f'token_addr{num}', contract_addr+'\n')
		elif key == 'thorchain_router':
			from mmgen.fileutil import write_data_to_file
			write_data_to_file(
				self.cfg,
				thorchain_router_addr_file,
				contract_addr + '\n',
				ask_overwrite = False,
				quiet = True)

		if mmgen_cmd == 'txdo':
			args += ['-k', keyfile]
		t = self.spawn('mmgen-'+mmgen_cmd, self.eth_opts + args)
		if mmgen_cmd == 'txcreate':
			t.written_to_file('transaction')
			ext = '[0,8000]{}.regtest.rawtx'.format('-α' if self.cfg.debug_utf8 else '')
			txfile = self.get_file_with_ext(ext, no_dot=True)
			t = self.spawn(
				'mmgen-txsign',
				self.eth_opts + ['--yes', '-k', keyfile, txfile], no_msg=True, no_passthru_opts=['coin'])
			self.txsign_ui_common(t, ni=True)

			txfile = txfile.replace('.rawtx', '.sigtx')
			t = self.spawn('mmgen-txsend',
				self.eth_opts + ['--wait', txfile], no_msg=True, no_passthru_opts=['coin'])

		self.txsend_ui_common(t,
			caller = mmgen_cmd,
			quiet  = mmgen_cmd == 'txdo' or not self.cfg.debug,
			contract_addr = contract_addr,
			bogus_send = False,
			wait = self.name == 'CmdTestEthBump')

		if key == 'Token':
			imsg(f'\nToken MM{num} deployed!')

		return t

	async def _token_deploy_math(self, *, num, mmgen_cmd='txdo'):
		return await self._token_deploy(
			num=num, key='SafeMath', gas=500_000, mmgen_cmd=mmgen_cmd)

	async def _token_deploy_owned(self, *, num):
		return await self._token_deploy(
			num=num, key='Owned', gas=1_000_000)

	async def _token_deploy_token(self, *, num):
		return await self._token_deploy(
			num=num, key='Token', gas=4_000_000, gas_price='7G')

	def _token_bal_check(self, *, pat):
		return self._bal_check(pat=pat, add_opts=['--token=MM1'])

	def _create_token_tx(self, *, cmd, fee, args, add_opts=[]):
		return self.txcreate_ui_common(
			self.spawn(
				f'mmgen-{cmd}',
				['--token=MM1', '-B', f'--fee={fee}'] + add_opts + args),
			inputs            = '1',
			input_sels_prompt = 'to spend from',
			caller            = cmd,
			file_desc         = 'Unsigned automount transaction')

	async def _token_transfer_ops(self, *, op, mm_idxs, amt=1000, sid=dfl_sid, token_addr=None):
		self.spawn(msg_only=True)
		from mmgen.tool.wallet import tool_cmd
		usr_mmaddrs = [f'{sid}:E:{i}' for i in mm_idxs]

		from mmgen.proto.eth.contract import ResolvedToken
		async def fund_user(rpc):
			for i in range(len(usr_mmaddrs)):
				tk = await ResolvedToken(
					self.cfg,
					self.proto,
					self.read_from_tmpfile(token_addr or f'token_addr{i+1}').strip(),
					rpc = rpc)
				imsg_r('\n' + await tk.info())
				imsg('dev token balance (pre-send): {}'.format(await tk.get_balance(dfl_devaddr)))
				imsg(f'Sending {amt} {self.proto.dcoin} to address {usr_addrs[i]} ({usr_mmaddrs[i]})')
				txid = await tk.transfer(
					from_addr = dfl_devaddr,
					to_addr   = usr_addrs[i],
					amt       = amt,
					key       = dfl_devkey,
					gas       = 120000,
					gasPrice  = self.proto.coin_amt(8, from_unit='Gwei'))
				rpc = await self.rpc
				imsg_r('Waiting for transaction receipt: ')
				for n in range(50): # long delay for txbump
					rx = await rpc.call('eth_getTransactionReceipt', '0x' + txid) # -> null if pending
					if rx:
						imsg('OK')
						break
					await asyncio.sleep(0.5)
					if n % 2:
						omsg_r('+')
				if not rx:
					die(1, 'tx receipt timeout exceeded')

		async def show_bals(rpc):
			for i in range(len(usr_mmaddrs)):
				tk = await ResolvedToken(
					self.cfg,
					self.proto,
					self.read_from_tmpfile(token_addr or f'token_addr{i+1}').strip(),
					rpc = rpc)
				imsg('Token: {}'.format(await tk.get_symbol()))
				imsg(f'dev token balance: {await tk.get_balance(dfl_devaddr)}')
				imsg('usr token balance: {} ({} {})'.format(
					await tk.get_balance(usr_addrs[i]),
					usr_mmaddrs[i],
					usr_addrs[i]))

		def gen_addr(addr):
			return tool_cmd(
				self.cfg, cmdname='gen_addr', proto=self.proto).gen_addr(addr, wallet=dfl_words_file)

		silence()
		usr_addrs = list(map(gen_addr, usr_mmaddrs))
		if op == 'show_bals':
			await show_bals(await self.rpc)
		elif op == 'fund_user':
			await fund_user(await self.rpc)
		end_silence()
		return 'ok'

class CmdTestEthdev(CmdTestEthdevMethods, CmdTestBase, CmdTestShared):
	'Ethereum transacting, token deployment and tracking wallet operations'
	networks = ('eth', 'etc')
	passthru_opts = ('coin', 'daemon_id', 'eth_daemon_id', 'http_timeout', 'rpc_backend')
	tmpdir_nums = [22]
	color = True
	menu_prompt = 'efresh balance:\b'
	input_sels_prompt = 'to spend from: '

	bals = lambda self, k: {
		'1': [  ('98831F3A:E:1', '123.456')],
		'2': [  ('98831F3A:E:1', '123.456'), ('98831F3A:E:11', '1.234')],
		'3': [  ('98831F3A:E:1', '123.456'), ('98831F3A:E:11', '1.234'), ('98831F3A:E:21', '2.345')],
		'4': [  ('98831F3A:E:1', '100'),
				('98831F3A:E:2', '23.45495'),
				('98831F3A:E:11', '1.234'),
				('98831F3A:E:21', '2.345')],
		'5': [  ('98831F3A:E:1', '100'),
				('98831F3A:E:2', '23.45495'),
				('98831F3A:E:11', '1.234'),
				('98831F3A:E:21', '2.345'),
				(burn_addr + r'\s+non-MMGen', amt1)],
		'8': [  ('98831F3A:E:1', '0'),
				('98831F3A:E:2', '23.45495'),
				('98831F3A:E:11', vbal1),
				('98831F3A:E:12', '99.99895'),
				('98831F3A:E:21', '2.345'),
				(burn_addr + r'\s+non-MMGen', amt1)],
		'9': [  ('98831F3A:E:1', '0'),
				('98831F3A:E:2', '23.45495'),
				('98831F3A:E:11', vbal1),
				('98831F3A:E:12', vbal2),
				('98831F3A:E:21', '2.345'),
				(burn_addr + r'\s+non-MMGen', amt1)],
		'10': [ ('98831F3A:E:1', '0'),
				('98831F3A:E:2', '23.0218'),
				('98831F3A:E:3', '0.4321'),
				('98831F3A:E:11', vbal1),
				('98831F3A:E:12', vbal2),
				('98831F3A:E:21', '2.345'),
				(burn_addr + r'\s+non-MMGen', amt1)]
	}[k]

	token_bals = lambda self, k: {
		'1': [  ('98831F3A:E:11', '1000', '1.234')],
		'2': [  ('98831F3A:E:11', '998.76544', vbal3),
				('98831F3A:E:12', '1.23456', '0')],
		'3': [  ('98831F3A:E:11', '110.654317776666555545', vbal1),
				('98831F3A:E:12', '1.23456', '0')],
		'4': [  ('98831F3A:E:11', '110.654317776666555545', vbal1),
				('98831F3A:E:12', '1.23456', '0'),
				(burn_addr + r'\s+non-MMGen', amt2, amt1)],
		'5': [  ('98831F3A:E:11', '110.654317776666555545', vbal1),
				('98831F3A:E:12', '1.23456', '99.99895'),
				(burn_addr + r'\s+non-MMGen', amt2, amt1)],
		'6': [  ('98831F3A:E:11', '110.654317776666555545', vbal1),
				('98831F3A:E:12', '0', vbal2),
				('98831F3A:E:13', '1.23456', '0'),
				(burn_addr + r'\s+non-MMGen', amt2, amt1)],
		'7': [  ('98831F3A:E:11', '67.444317776666555545', vbal9),
				('98831F3A:E:12', '43.21', vbal2),
				('98831F3A:E:13', '1.23456', '0'),
				(burn_addr + r'\s+non-MMGen', amt2, amt1)]
	}[k]

	token_bals_getbalance = lambda self, k: {
		'1': (vbal4, '999777.12345689012345678'),
		'2': ('111.888877776666555545', '888.111122223333444455')
	}[k]

	cmd_group_in = (
		('setup',             f'dev mode tests for coin {coin} (start daemon)'),
		('subgroup.misc',     []),
		('subgroup.init',     []),
		('subgroup.msg',      ['init']),
		('subgroup.main',     ['init']),
		('subgroup.contract', ['main']),
		('subgroup.token',    ['contract']),
		('subgroup.twexport', ['token']),
		('subgroup.cached',   ['token']),
		('subgroup.view',     ['cached']),
		('subgroup.label',    ['cached']),
		('subgroup.remove',   ['cached']),
		('stop',              'stopping daemon'),
	)
	cmd_subgroups = {
	'misc': (
		'miscellaneous commands',
		('daemon_version', 'mmgen-tool daemon_version'),
	),
	'init': (
		'initializing wallets',
		('wallet_upgrade1',         'upgrading the tracking wallet (v1 -> v2)'),
		('wallet_upgrade2',         'upgrading the tracking wallet (v2 -> v3)'),
		('addrgen',                 'generating addresses'),
		('addrimport',              'importing addresses'),
		('addrimport_devaddr',      'importing the dev address'),
		('addrimport_reth_devaddr', 'importing the reth dev address'),
		('fund_devaddr',            'funding the dev address'),
		('del_reth_devaddr',        'deleting the reth dev address'),
		('cli_dev_balance',         'mmgen-cli eth_getBalance'),
	),
	'msg': (
		'message signing',
		('msgsign_chk',          "signing a message (low-level, check against 'eth_sign' RPC call)"),
		('msgcreate',            'creating a message file'),
		('msgsign',              'signing the message file'),
		('msgverify',            'verifying the message file'),
		('msgexport',            'exporting the message file data to JSON for third-party verifier'),
		('msgverify_export',     'verifying the exported JSON data'),

		('msgcreate_raw',        'creating a message file (--msghash-type=raw)'),
		('msgsign_raw',          'signing the message file (msghash_type=raw)'),
		('msgverify_raw',        'verifying the message file (msghash_type=raw)'),
		('msgexport_raw',        'exporting the message file data to JSON (msghash_type=raw)'),
		('msgverify_export_raw', 'verifying the exported JSON data (msghash_type=raw)'),
	),
	'main': (
		'creating, signing, sending and bumping Ethereum transactions',
		('txcreate1',            'creating a transaction (spend from dev address to address :1)'),
		('txview1_raw',          'viewing the raw transaction'),
		('txsign1',              'signing the transaction'),
		('txview1_sig',          'viewing the signed transaction'),
		('tx_status0_bad',       'getting the transaction status'),
		('txsign1_ni',           'signing the transaction (non-interactive)'),

		('etherscan_server_start','starting the Etherscan server'),
		('txsend_etherscan_test','sending the transaction via Etherscan (simulation, with --test)'),
		('txsend_etherscan',     'sending the transaction via Etherscan (simulation)'),
		('etherscan_server_stop','stopping the Etherscan server'),

		('txsend1',              'sending the transaction'),
		('bal1',                 f'the {coin} balance'),

		('txcreate2',            'creating a transaction (spend from dev address to address :11)'),
		('txsign2',              'signing the transaction'),
		('txsend2',              'sending the transaction'),
		('bal2',                 f'the {coin} balance'),

		('txcreate3',            'creating a transaction (spend from dev address to address :21)'),
		('txsign3',              'signing the transaction'),
		('txsend3',              'sending the transaction'),
		('bal3',                 f'the {coin} balance'),

		('tx_status1',           'getting the transaction status'),

		('txcreate4',            'creating a transaction (spend from MMGen address, low TX fee)'),
		('txbump',               'bumping the transaction fee'),

		('txsign4',              'signing the transaction'),
		('txsend4',              'sending the transaction'),
		('tx_status1a',          'getting the transaction status'),
		('bal4',                 f'the {coin} balance'),

		('txcreate5',            'creating a transaction (fund burn address)'),
		('txsign5',              'signing the transaction'),
		('txsend5',              'sending the transaction'),

		('addrimport_burn_addr', 'importing burn address'),
		('bal5',                 f'the {coin} balance'),

		('add_comment1',         'adding a UTF-8 label (zh)'),
		('chk_comment1',         'checking the label'),
		('add_comment2',         'adding a UTF-8 label (lat+cyr+gr)'),
		('chk_comment2',         'checking the label'),
		('remove_comment',       'removing the label'),
	),
	'contract': (
		'creating and deploying ERC20 tokens',
		('token_compile1',  'compiling ERC20 token #1'),
		('token_deploy1a',  'deploying ERC20 token #1 (SafeMath)'),
		('token_deploy1b',  'deploying ERC20 token #1 (Owned)'),
		('token_deploy1c',  'deploying ERC20 token #1 (Token)'),

		('tx_status2',      'getting the transaction status'),
		('bal6',            f'the {coin} balance'),

		('token_compile2',  'compiling ERC20 token #2'),
		('token_deploy2a',  'deploying ERC20 token #2 (SafeMath)'),
		('token_deploy2b',  'deploying ERC20 token #2 (Owned)'),
		('token_deploy2c',  'deploying ERC20 token #2 (Token)'),
	),
	'token': (
		'creating, signing, sending and bumping ERC20 token transactions',

		('token_fund_users',           'transferring token funds from dev to user'),
		('token_user_bals',            'show balances after transfer'),
		('token_addrgen',              'generating token addresses'),
		('token_addrimport_badaddr1',  'importing token addresses (no token address)'),
		('token_addrimport_badaddr2',  'importing token addresses (bad token address)'),
		('token_addrimport_addr1',     'importing token addresses using token address (MM1)'),
		('token_addrimport_addr2',     'importing token addresses using token address (MM2)'),
		('token_addrimport_batch',     'importing token addresses (dummy batch mode) (MM1)'),
		('token_addrimport_sym',       'importing token addresses using token symbol (MM2)'),

		('bal7',                       f'the {coin} balance'),
		('token_bal1',                 f'the {coin} balance and token balance'),

		('token_txcreate1',            'creating a token transaction'),
		('token_txview1_raw',          'viewing the raw transaction'),
		('token_txsign1',              'signing the transaction'),
		('token_txsend1',              'sending the transaction'),
		('token_txview1_sig',          'viewing the signed transaction'),
		('tx_status3',                 'getting the transaction status'),
		('token_bal2',                 f'the {coin} balance and token balance'),

		('token_txcreate2',            'creating a token transaction (to burn address)'),
		('token_txbump',               'bumping the transaction fee'),

		('token_txsign2',              'signing the transaction'),
		('token_txsend2',              'sending the transaction'),
		('token_bal3',                 f'the {coin} balance and token balance'),

		('del_devaddr',                'deleting the dev address'),

		('bal1_getbalance',            f'the {coin} balance (getbalance)'),

		('addrimport_token_burn_addr', 'importing the token burn address'),

		('token_bal4',                 f'the {coin} balance and token balance'),
		('token_bal_getbalance',       'the token balance (getbalance)'),

		('txcreate_noamt',             'creating a transaction (full amount send)'),
		('txsign_noamt',               'signing the transaction'),
		('txsend_noamt',               'sending the transaction'),

		('bal8',                       f'the {coin} balance'),
		('token_bal5',                 'the token balance'),

		('token_txcreate_noamt',       'creating a token transaction (full amount send)'),
		('token_txsign_noamt',         'signing the transaction'),
		('token_txsend_noamt',         'sending the transaction'),

		('bal9',                       f'the {coin} balance'),
		('token_bal6',                 'the token balance'),

		('listaddresses1',             'listaddresses'),
		('listaddresses2',             'listaddresses minconf=3'),
		('listaddresses3',             'listaddresses sort=age (ignored)'),
		('listaddresses4',             'listaddresses showempty=1 sort=age (ignored)'),

		('token_listaddresses1',       'listaddresses --token=mm1'),
		('token_listaddresses2',       'listaddresses --token=mm1 showempty=1'),
	),
	'twexport': (
		'exporting and importing tracking wallet to JSON',
		('twexport_noamt',       'exporting the tracking wallet (include_amts=0)'),
		('twmove',               'moving the tracking wallet'),
		('twimport',             'importing the tracking wallet'),
		('twview7',              'twview (cached_balances=1)'),
		('twview8',              'twview'),
		('twexport',             'exporting the tracking wallet'),
		('tw_chktotal',          'checking total value in tracking wallet dump'),
		('twmove',               'moving the tracking wallet'),
		('twimport',             'importing the tracking wallet'),
		('twcompare',            'comparing imported tracking wallet with original'),
		('edit_json_twdump',     'editing the tracking wallet JSON dump'),
		('twmove',               'moving the tracking wallet'),
		('twimport_nochksum',    'importing the edited tracking wallet JSON dump (ignore_checksum=1)'),

		('token_listaddresses3', 'listaddresses --token=mm1 showempty=1'),
		('token_listaddresses4', 'listaddresses --token=mm2 showempty=1'),
		('twview9',              'twview (check balance)'),
	),
	'cached': (
		'creating and sending transactions using cached balances',
		('twview_cached_balances',          'twview (cached balances)'),
		('token_twview_cached_balances',    'token twview (cached balances)'),
		('txcreate_cached_balances',        'txcreate (cached balances)'),
		('token_txcreate_cached_balances',  'token txcreate (cached balances)'),

		('txdo_cached_balances',            'txdo (cached balances)'),
		('txcreate_refresh_balances',       'refreshing balances'),
		('bal10',                           f'the {coin} balance'),

		('token_txdo_cached_balances',      'token txdo (cached balances)'),
		('token_txcreate_refresh_balances', 'refreshing token balances'),
		('token_bal7',                      'the token balance'),
	),
	'view': (
		'viewing addresses and unspent outputs',
		('twview1',       'twview'),
		('twview2',       'twview wide=1'),
		('twview3',       'twview wide=1 sort=age (ignored)'),
		('twview4',       'twview wide=1 minconf=15'),
		('twview5',       'twview wide=1 minconf=0'),
		('token_twview1', 'twview --token=mm1'),
		('token_twview2', 'twview --token=mm1 wide=1'),
		('token_twview3', 'twview --token=mm1 wide=1 sort=age (ignored)'),
	),
	'label': (
		'creating, editing and removing labels',
		('edit_comment1',       f'adding label to addr #{del_addrs[0]} in {coin} tracking wallet (zh)'),
		('edit_comment2',       f'editing label for addr #{del_addrs[0]} in {coin} tracking wallet (zh)'),
		('edit_comment3',       f'adding label to addr #{del_addrs[1]} in {coin} tracking wallet (lat+cyr+gr)'),
		('edit_comment4',       f'removing label from addr #{del_addrs[0]} in {coin} tracking wallet'),
		('token_edit_comment1', f'adding label to addr #{del_addrs[0]} in {coin} token tracking wallet'),
	),
	'remove': (
		'removing addresses from tracking wallet',
		('remove_addr1',       f'removing addr #{del_addrs[0]} from {coin} tracking wallet'),
		('twview6',            'twview (balance reduced after address removal)'),
		('remove_addr2',       f'removing addr #{del_addrs[1]} from {coin} tracking wallet'),
		('token_remove_addr1', f'removing addr #{del_addrs[0]} from {coin} token tracking wallet'),
		('token_remove_addr2', f'removing addr #{del_addrs[1]} from {coin} token tracking wallet'),
	),
	}

	def __init__(self, cfg, trunner, cfgs, spawn):
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)
		if trunner is None:
			return

		global coin
		coin = cfg.coin

		self.eth_opts         = [f'--outdir={self.tmpdir}', '--regtest=1', '--quiet']
		self.eth_opts_noquiet = [f'--outdir={self.tmpdir}', '--regtest=1']

		from mmgen.protocol import init_proto
		self.proto = init_proto(cfg, network_id=self.proto.coin+'_rt', need_amt=True)

		from mmgen.daemon import CoinDaemon
		self.daemon = CoinDaemon(cfg, network_id=self.proto.coin+'_rt', test_suite=True)

		if self.daemon.id == 'reth':
			global reth_devkey, reth_devaddr
			reth_devkey, reth_devaddr = get_reth_dev_keypair(cfg)
			write_to_file(joinpath(self.tmpdir, reth_devkey_fn), reth_devkey+'\n')

		set_vbals(self.daemon.id)

		self.using_solc = check_solc_ver()
		if not self.using_solc:
			omsg(yellow('Using precompiled contract data'))

		omsg(orange(f'Coin daemon {self.daemon.id!r} selected'))

		self.genesis_fn = joinpath(self.tmpdir, 'genesis.json')
		self.keystore_dir = os.path.relpath(joinpath(self.daemon.datadir, 'keystore'))

		write_to_file(
			joinpath(self.tmpdir, dfl_devkey_fn),
			dfl_devkey+'\n')

		self.message = 'attack at dawn'
		self.spawn_env['MMGEN_BOGUS_SEND'] = ''

		TestProxy(self, cfg)

	@property
	async def rpc(self):
		from mmgen.rpc import rpc_init
		return await rpc_init(self.cfg, self.proto)

	def mining_delay(self): # workaround for mining race condition in dev mode
		if self.daemon.id == 'reth':
			time.sleep(1)

	async def setup(self):
		self.spawn(msg_only=True)

		d = self.daemon

		if not self.using_solc:
			srcdir = os.path.join(self.tr.repo_root, 'test', 'ref', 'ethereum', 'bin')
			from shutil import copytree
			for _ in ('mm1', 'mm2'):
				copytree(os.path.join(srcdir, _), os.path.join(self.tmpdir, _))

		if d.id in ('geth', 'erigon'):
			self.genesis_setup(d)
			set_vt100()

		if d.id == 'erigon':
			self.write_to_tmpfile('signer_key', self.keystore_data['key']+'\n')
			d.usr_coind_args = [
				'--miner.sigfile={}'.format(os.path.join(self.tmpdir, 'signer_key')),
				'--miner.etherbase={}'.format(self.keystore_data['address'])]

		if d.id in ('geth', 'erigon'):
			imsg('  {:19} {}'.format('Cmdline:', ' '.join(e for e in d.start_cmd if not 'verbosity' in e)))

		if not self.cfg.no_daemon_autostart:
			if not d.id in ('geth', 'erigon'):
				d.stop(silent=True)
				d.remove_datadir()
			self.init_block_period()
			d.start(silent=self.tr.quiet)
			rpc = await self.rpc
			imsg(f'Daemon: {rpc.daemon.coind_name} v{rpc.daemon_version_str}')

		return 'ok'

	def init_block_period(self):
		pass

	@property
	def keystore_data(self):
		if not hasattr(self, '_keystore_data'):

			wallet_fn = os.path.join( self.keystore_dir, os.listdir(self.keystore_dir)[0])

			from mmgen.proto.eth.util import decrypt_geth_keystore
			key = decrypt_geth_keystore(
				cfg       = self.cfg,
				wallet_fn = wallet_fn,
				passwd = b'')

			with open(wallet_fn) as fh:
				res = json.loads(fh.read())

			res.update( { 'key': key.hex()})
			self._keystore_data = res

		return self._keystore_data

	def genesis_setup(self, d):

		def make_key():
			pwfile = joinpath(self.tmpdir, 'account_passwd')
			write_to_file(pwfile, '')
			run(['rm', '-rf', self.keystore_dir])
			cmd = f'geth account new --password={pwfile} --lightkdf --keystore {self.keystore_dir}'
			cp = run(cmd.split(), stdout=PIPE, stderr=PIPE)
			if cp.returncode:
				die(1, cp.stderr.decode())

		def make_genesis(signer_addr, prealloc_addr):
			return {
				'config': {
					'chainId': 1337, # TODO: replace constant with var
					'homesteadBlock': 0,
					'eip150Block': 0,
					'eip155Block': 0,
					'eip158Block': 0,
					'byzantiumBlock': 0,
					'constantinopleBlock': 0,
					'petersburgBlock': 0,
					'istanbulBlock': 0,
					'muirGlacierBlock': 0,
					'berlinBlock': 0,
					'londonBlock': 0,
					'arrowGlacierBlock': 0,
					'grayGlacierBlock': 0,
					'shanghaiTime': 0,
					'terminalTotalDifficulty': 0,
					'terminalTotalDifficultyPassed': True,
					'isDev': True
				},
				'nonce': '0x0',
				'timestamp': '0x0',
				'extraData': '0x',
				'gasLimit': '0xaf79e0',
				'difficulty': '0x0',
				'mixHash': '0x0000000000000000000000000000000000000000000000000000000000000000',
				'coinbase': '0x0000000000000000000000000000000000000000',
				'number': '0x0',
				'gasUsed': '0x0',
				'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000000',
				'baseFeePerGas': '0x3b9aca00',
				'excessBlobGas': None,
				'blobGasUsed': None,
				'alloc': {
					prealloc_addr: { 'balance': hex(prealloc_amt.toWei())}
				}
			}

		def init_genesis(fn):
			cmd = f'{d.exec_fn} init --datadir {d.datadir} {fn}'
			cp = run( cmd.split(), stdout=PIPE, stderr=PIPE)
			if cp.returncode:
				die(1, cp.stderr.decode())

		d.stop(quiet=True)
		d.remove_datadir()

		imsg(cyan('Initializing Genesis Block:'))

		prealloc_amt = self.proto.coin_amt('1_000_000_000')

		make_key()
		signer_addr = self.keystore_data['address']
		self.write_to_tmpfile('signer_addr', signer_addr + '\n')

		imsg(f'  Keystore:           {self.keystore_dir}')
		imsg(f'  Signer key:         {self.keystore_data["key"]}')
		imsg(f'  Signer address:     {signer_addr}')
		imsg(f'  Faucet:             {dfl_devaddr} ({prealloc_amt} ETH)')
		imsg(f'  Genesis block data: {self.genesis_fn}')

		genesis_data = make_genesis(signer_addr, dfl_devaddr)
		write_to_file( self.genesis_fn, json.dumps(genesis_data, indent='  ')+'\n')
		init_genesis(self.genesis_fn)

	def daemon_version(self):
		t = self.spawn('mmgen-tool', self.eth_opts + ['daemon_version'])
		t.expect('version')
		return t

	def cli_dev_balance(self):
		self.mining_delay()
		t = self.spawn(
			'mmgen-cli',
			[f'--coin={self.proto.coin}', '--regtest=1', 'eth_getBalance', '0x'+dfl_devaddr, 'latest'])
		if self.daemon.id == 'geth':
			t.expect('0x33b2e3c91ec0e9113986000')
		elif self.daemon.id == 'reth':
			t.expect('0xd3c21bab45fb313f0000')
		return t

	async def _wallet_upgrade(self, src_fn, expect1, expect2=None):
		if self.proto.coin == 'ETC':
			msg(f'skipping test {self.test_name!r} for ETC')
			return 'skip'
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(self.cfg, self.proto, no_wallet_init=True)
		from_fn = Path(ref_dir) / 'ethereum' / src_fn
		bak_fn = twctl.tw_dir / f'upgraded-{src_fn}'
		twctl.tw_dir.mkdir(mode=0o750, parents=True, exist_ok=True)
		dest = shutil.copy2(from_fn, twctl.tw_path)
		assert dest == twctl.tw_path, f'{dest} != {twctl.tw_path}'
		t = self.spawn('mmgen-tool', self.eth_opts + ['twview'])
		t.expect(expect1)
		if expect2:
			t.expect(expect2)
		t.read()
		twctl.tw_path.rename(bak_fn)
		return t

	async def wallet_upgrade1(self):
		return await self._wallet_upgrade('tracking-wallet-v1.json', 'accounts field', 'network field')

	async def wallet_upgrade2(self):
		return await self._wallet_upgrade('tracking-wallet-v2.json', 'token params field', 'network field')

	def addrgen(self):
		return self._addrgen()

	def addrimport(self):
		return self._addrimport()

	def _addrimport_one_addr(self, addr=None, extra_args=[]):
		t = self.spawn('mmgen-addrimport', ['--regtest=1', '--quiet', f'--address={addr}'] + extra_args)
		t.expect('OK')
		return t

	def addrimport_devaddr(self):
		return self._addrimport_one_addr(addr=dfl_devaddr)

	def addrimport_reth_devaddr(self):
		if not self.daemon.id == 'reth':
			return 'silent'
		return self._addrimport_one_addr(addr=reth_devaddr)

	def addrimport_burn_addr(self):
		return self._addrimport_one_addr(addr=burn_addr)

	def txcreate(
			self,
			args            = [],
			menu            = [],
			acct            = '1',
			caller          = 'txcreate',
			interactive_fee = '50G',
			fee_info_data   = ('0.00105', '50'),
			no_read         = False,
			print_listing   = True,
			bad_input_sels  = True,
			tweaks          = []):
		fee_info_pat = r'\D{}\D.*{c} .*\D{}\D.*gas price in Gwei'.format(*fee_info_data, c=self.proto.coin)
		t = self.spawn(f'mmgen-{caller}', self.eth_opts + ['-B'] + args)
		if print_listing:
			t.expect(r'add \[l\]abel, .*?:.', 'p', regex=True)
			t.written_to_file('Account balances listing')
		t = self.txcreate_ui_common(
			t,
			menu              = menu,
			caller            = caller,
			input_sels_prompt = 'to spend from',
			inputs            = acct,
			file_desc         = 'transaction',
			bad_input_sels    = bad_input_sels,
			interactive_fee   = interactive_fee,
			fee_info_pat      = fee_info_pat,
			fee_desc          = 'transaction fee or gas price',
			add_comment       = tx_comment_jp,
			tweaks            = tweaks)
		if not no_read:
			t.read()
		return t

	def txsign(self, ni=False, ext='{}.regtest.rawtx', add_args=[], dev_send=False, has_label=True):
		ext = ext.format('-α' if self.cfg.debug_utf8 else '')
		keyfile = joinpath(self.tmpdir, dfl_devkey_fn)
		txfile = self.get_file_with_ext(ext, no_dot=True)
		t = self.spawn(
			'mmgen-txsign',
				self.eth_opts
				+ ['--rpc-host=bad_host'] # ETH signing must work without RPC
				+ ([], ['--yes'])[ni]
				+ ([f'--keys-from-file={keyfile}'] if dev_send else [])
				+ add_args
				+ [txfile, dfl_words_file],
			no_passthru_opts = ['coin'])
		return self.txsign_ui_common(t, ni=ni, has_label=has_label)

	def txsend(
			self,
			ext          = '{}.regtest.sigtx',
			add_args     = [],
			test         = False,
			return_early = False,
			has_label    = True,
			env          = {}):
		ext = ext.format('-α' if self.cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		t = self.spawn(
			'mmgen-txsend',
			self.eth_opts + add_args + [txfile],
			no_passthru_opts = ['coin'],
			env = env)
		if return_early:
			return t
		self.txsend_ui_common(
			t,
			quiet      = not self.cfg.debug,
			bogus_send = False,
			test       = test,
			has_label  = has_label)
		return t

	def txview(self, ext_fs):
		ext = ext_fs.format('-α' if self.cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		return self.spawn('mmgen-tool', ['--verbose', 'txview', txfile])

	def fund_devaddr(self):
		"""
		For Reth, fund the default (Parity) dev address from the Reth dev address
		For the others, send a junk TX to keep block counts equal for all daemons
		"""
		dt = namedtuple('data', ['devkey_fn', 'dest', 'amt'])
		if self.daemon.id == 'reth':
			d = dt(reth_devkey_fn, dfl_devaddr, Decimal(reth_dev_amt) - Decimal('0.01'))
		else:
			d = dt(dfl_devkey_fn, burn_addr2, '1')
		t = self.txcreate(
			args    = self.eth_opts_noquiet + [
				f'--keys-from-file={joinpath(self.tmpdir, d.devkey_fn)}',
				f'{d.dest},{d.amt}',
			],
			menu    = ['a', 'r'],
			caller  = 'txdo',
			acct    = '1',
			no_read = True)
		self._do_confirm_send(t, quiet=not self.cfg.debug, sure=False)
		t.read()
		self.get_file_with_ext('sigtx', delete_all=True)
		return t

	def del_reth_devaddr(self):
		if not self.daemon.id == 'reth':
			return 'silent'
		return self._del_addr(reth_devaddr)

	def txcreate1(self):
		# include one invalid keypress 'X' -- see EthereumTwUnspentOutputs.key_mappings
		menu = ['a', 'd', 'r', 'M', 'X', 'e', 'm', 'm']
		args = ['98831F3A:E:1,123.456']
		return self.txcreate(args=args, menu=menu, acct='1', tweaks=['confirm_non_mmgen'])
	def txview1_raw(self):
		return self.txview(ext_fs='{}.regtest.rawtx')
	def txsign1(self):
		return self.txsign(add_args=['--use-internal-keccak-module'], dev_send=True)
	def tx_status0_bad(self):
		return self.tx_status(
			ext        = '{}.regtest.sigtx',
			expect_str = 'neither in mempool nor blockchain',
			exit_val   = 1)
	def txsign1_ni(self):
		return self.txsign(ni=True, dev_send=True)

	def etherscan_server_start(self):
		if self.proto.coin == 'ETC':
			return 'skip'
		self.spawn(msg_only=True)
		etherscan_server.start()
		return 'ok'

	def txsend_etherscan_test(self):
		if self.proto.coin == 'ETC':
			return 'skip'
		return self.txsend(add_args=['--tx-proxy=ether', '--test'], test='tx_proxy')

	def txsend_etherscan(self):
		if self.proto.coin == 'ETC':
			return 'skip'
		return self.txsend(add_args=['--tx-proxy=ethersc', f'--proxy=localhost:{TestProxy.port}'])

	def etherscan_server_stop(self):
		if self.proto.coin == 'ETC':
			return 'skip'
		self.spawn(msg_only=True)
		etherscan_server.stop()
		return 'ok'

	def txsend1(self):
		return self.txsend()
	def txview1_sig(self): # do after send so that TxID is displayed
		return self.txview(ext_fs='{}.regtest.sigtx')
	def bal1(self):
		return self.bal(n='1')

	def txcreate2(self):
		args = ['98831F3A:E:11,1.234']
		return self.txcreate(args=args, acct='10', tweaks=['confirm_non_mmgen'])
	def txsign2(self):
		return self.txsign(ni=True, ext='1.234,50000]{}.regtest.rawtx', dev_send=True)
	def txsend2(self):
		return self.txsend(ext='1.234,50000]{}.regtest.sigtx')
	def bal2(self):
		return self.bal(n='2')

	def txcreate3(self):
		args = ['98831F3A:E:21,2.345']
		return self.txcreate(args=args, acct='10', tweaks=['confirm_non_mmgen'])
	def txsign3(self):
		return self.txsign(ni=True, ext='2.345,50000]{}.regtest.rawtx', dev_send=True)
	def txsend3(self):
		return self.txsend(ext='2.345,50000]{}.regtest.sigtx')
	def bal3(self):
		return self.bal(n='3')

	def tx_status(self, ext, *, expect_str, expect_str2='', add_opts=[], exit_val=0):
		self.mining_delay()
		ext = ext.format('-α' if self.cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext, no_dot=True)
		t = self.spawn(
			'mmgen-txsend',
			self.eth_opts + add_opts + ['--status', txfile],
			no_passthru_opts = ['coin'],
			exit_val = exit_val)
		t.expect(expect_str)
		if expect_str2:
			t.expect(expect_str2)
		if '--verbose' in add_opts:
			t.expect('view: ', 'n')
		return t

	def tx_status1(self):
		return self.tx_status(
			ext        = '2.345,50000]{}.regtest.sigtx',
			add_opts   = ['--verbose'],
			expect_str = 'has 1 confirmation')

	def tx_status1a(self):
		return self.tx_status(ext='2.345,50000]{}.regtest.sigtx', expect_str='has 2 confirmations')

	async def msgsign_chk(self): # NB: Geth only!

		def create_signature_mmgen():
			key = self.keystore_data['key']
			imsg(f'Key:       {key}')
			from mmgen.proto.eth.util import ec_sign_message_with_privkey
			return ec_sign_message_with_privkey(self.cfg, self.message, bytes.fromhex(key), 'eth_sign')

		async def create_signature_rpc():
			addr = self.read_from_tmpfile('signer_addr').strip()
			imsg(f'Address:   {addr}')
			rpc = await self.rpc
			return await rpc.call(
				'eth_sign',
				'0x' + addr,
				'0x' + self.message.encode().hex())

		if not self.daemon.id == 'geth':
			return 'skip'

		self.spawn(msg_only=True)

		sig = '0x' + create_signature_mmgen()
		sig_chk = await create_signature_rpc()

		# Compare signatures
		imsg(f'Message:   {self.message}')
		imsg(f'Signature: {sig}')
		if sig != sig_chk:
			self.tr.warn('MMGen signature doesn’t match reference signature from Geth daemon')

		return 'ok'

	def msgcreate(self, add_args=[]):
		t = self.spawn('mmgen-msg', self.eth_opts_noquiet + add_args + ['create', self.message, '98831F3A:E:1'])
		t.written_to_file('Unsigned message data')
		return t

	def msgsign(self):
		fn = get_file_with_ext(self.tmpdir, 'rawmsg.json')
		t = self.spawn('mmgen-msg', self.eth_opts_noquiet + ['sign', fn, dfl_words_file])
		t.written_to_file('Signed message data')
		return t

	def msgverify(self, fn=None, msghash_type='eth_sign'):
		fn = fn or get_file_with_ext(self.tmpdir, 'sigmsg.json')
		t = self.spawn('mmgen-msg', self.eth_opts_noquiet + ['verify', fn])
		t.expect(msghash_type)
		t.expect('1 signature verified')
		return t

	def msgexport(self):
		fn = get_file_with_ext(self.tmpdir, 'sigmsg.json')
		t = self.spawn('mmgen-msg', self.eth_opts_noquiet + ['export', fn])
		t.written_to_file('Signature data')
		return t

	def msgverify_export(self):
		return self.msgverify(
			fn = os.path.join(self.tmpdir, 'signatures.json'))

	def msgcreate_raw(self):
		get_file_with_ext(self.tmpdir, 'rawmsg.json', delete_all=True)
		return self.msgcreate(add_args=['--msghash-type=raw'])

	def msgsign_raw(self):
		get_file_with_ext(self.tmpdir, 'sigmsg.json', delete_all=True)
		return self.msgsign()

	def msgverify_raw(self):
		return self.msgverify(msghash_type='raw')

	def msgexport_raw(self):
		get_file_with_ext(self.tmpdir, 'signatures.json', no_dot=True, delete_all=True)
		return self.msgexport()

	def msgverify_export_raw(self):
		return self.msgverify(
			fn = os.path.join(self.tmpdir, 'signatures.json'),
			msghash_type = 'raw')

	def txcreate4(self):
		return self.txcreate(
			args             = ['98831F3A:E:2,23.45495'],
			acct             = '1',
			interactive_fee  = '40G',
			fee_info_data    = ('0.00084', '40'))

	def txbump(self):
		return self._txbump(fee='50G', ext=',40000]{}.regtest.rawtx')

	def txsign4(self):
		return self.txsign(ext='.45495,50000]{}.regtest.rawtx', add_args=['--no-quiet', '--no-yes'])
	def txsend4(self):
		return self.txsend(ext='.45495,50000]{}.regtest.sigtx')
	def bal4(self):
		return self.bal(n='4')

	def txcreate5(self):
		args = [burn_addr + ','+amt1]
		return self.txcreate(args=args, acct='10', tweaks=['confirm_non_mmgen'])
	def txsign5(self):
		return self.txsign(ni=True, ext=amt1+',50000]{}.regtest.rawtx', dev_send=True)
	def txsend5(self):
		return self.txsend(ext=amt1+',50000]{}.regtest.sigtx')
	def bal5(self):
		return self.bal(n='5')

	def bal(self, n):
		self.mining_delay()
		t = self.spawn('mmgen-tool', self.eth_opts + ['twview', 'wide=1'])
		text = t.read(strip_color=True)
		for addr, amt in self.bals(n):
			pat = r'\D{}\D.*\D{}\D'.format(addr, amt.replace('.', r'\.'))
			assert re.search(pat, text), pat
		ss = f'Total {self.proto.coin}:'
		assert re.search(ss, text), ss
		return t

	def token_bal(self, n=None):
		self.mining_delay()
		t = self.spawn('mmgen-tool', self.eth_opts + ['--token=mm1', 'twview', 'wide=1'])
		text = t.read(strip_color=True)
		for addr, _amt1, _amt2 in self.token_bals(n):
			pat = fr'{addr}\b.*\D{_amt1}\D.*\b{_amt2}\D'
			assert re.search(pat, text), pat
		ss = 'Total MM1:'
		assert re.search(ss, text), ss
		return t

	def bal_getbalance(self, sid, idx, etc_adj=False, extra_args=[]):
		bal1 = self.token_bals_getbalance(idx)[0]
		bal2 = self.token_bals_getbalance(idx)[1]
		bal1 = Decimal(bal1)
		t = self.spawn('mmgen-tool', self.eth_opts + extra_args + ['getbalance'])
		t.expect(rf'{sid}:.*'+str(bal1), regex=True)
		t.expect(r'Non-MMGen:.*'+bal2, regex=True)
		total = strip_ansi_escapes(t.expect_getend(rf'TOTAL {self.proto.coin}')).split()[0]
		assert Decimal(bal1) + Decimal(bal2) == Decimal(total)
		return t

	def add_comment(self, comment, addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_opts + ['add_label', addr, comment])
		t.expect('Added label.*in tracking wallet', regex=True)
		return t

	def chk_comment(self, comment_pat, addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_opts + ['listaddresses', 'all_labels=1'])
		t.expect(fr'{addr}\b.*{comment_pat}', regex=True)
		return t

	def add_comment1(self):
		return self.add_comment(comment=tw_comment_zh)
	def chk_comment1(self):
		return self.chk_comment(comment_pat=tw_comment_zh[:3])
	def add_comment2(self):
		return self.add_comment(comment=tw_comment_lat_cyr_gr)
	def chk_comment2(self):
		return self.chk_comment(comment_pat=tw_comment_lat_cyr_gr[:3])

	def remove_comment(self, addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_opts + ['remove_label', addr])
		t.expect('Removed label.*in tracking wallet', regex=True)
		return t

	def token_compile(self, token_data={}):
		odir = joinpath(self.tmpdir, token_data['symbol'].lower())
		if not self.using_solc:
			imsg(f'Using precompiled contract data in {odir}')
			return 'skip' if os.path.exists(odir) else False
		self.spawn(msg_only=True)
		cmd_args = [f'--{k}={v}' for k, v in list(token_data.items())]
		imsg("Compiling solidity token contract '{}' with 'solc'".format(token_data['symbol']))
		try:
			os.mkdir(odir)
		except:
			pass
		cmd = [
			'python3',
			'scripts/create-token.py',
			'--coin=' + self.proto.coin,
			'--outdir=' + odir
		] + cmd_args + [self.proto.checksummed_addr(dfl_devaddr)]
		imsg('Executing: {}'.format(' '.join(cmd)))
		cp = run(cmd, stdout=DEVNULL, stderr=PIPE)
		if cp.returncode != 0:
			rmsg('solc failed with the following output:')
			die(2, cp.stderr.decode())
		imsg('ERC20 token {!r} compiled'.format(token_data['symbol']))
		return 'ok'

	def token_compile1(self):
		token_data = {'name':'MMGen Token 1', 'symbol':'MM1', 'supply':10**26, 'decimals':18}
		return self.token_compile(token_data)

	def token_compile2(self):
		token_data = {'name':'MMGen Token 2', 'symbol':'MM2', 'supply':10**18, 'decimals':10}
		return self.token_compile(token_data)

	async def token_deploy1a(self):
		return await self._token_deploy_math(num=1)

	async def token_deploy1b(self):
		return await self._token_deploy_owned(num=1)

	async def token_deploy1c(self):
		return await self._token_deploy_token(num=1)

	async def token_deploy2a(self): # test create, sign, send:
		return await self._token_deploy_math(num=2, mmgen_cmd='txcreate')

	async def token_deploy2b(self):
		return await self._token_deploy_owned(num=2)

	async def token_deploy2c(self):
		return await self._token_deploy_token(num=2)

	def tx_status2(self):
		return self.tx_status(
			ext        = self.proto.coin+'[0,7000]{}.regtest.sigtx',
			expect_str = 'successfully executed')

	def bal6(self):
		return self.bal5()

	def token_fund_users(self):
		return self._token_transfer_ops(op='fund_user', mm_idxs=[11, 21])

	def token_user_bals(self):
		return self._token_transfer_ops(op='show_bals', mm_idxs=[11, 21])

	def token_addrgen(self):
		return self._token_addrgen(mm_idxs=[11, 21], naddrs=3)

	def token_addrimport_badaddr1(self):
		t = self._addrimport(
			ext       = '[11-13]{}.regtest.addrs',
			add_args  = ['--token=abc'],
			bad_input = True,
			exit_val  = 2)
		t.expect('could not be resolved')
		return t

	def token_addrimport_badaddr2(self):
		t = self._addrimport(
			ext       = '[11-13]{}.regtest.addrs',
			add_args  = ['--token='+'00deadbeef'*4],
			bad_input = True,
			exit_val  = 2)
		t.expect('could not be resolved')
		return t

	def token_addrimport_addr1(self):
		return self._token_addrimport('token_addr1', '11-13', expect='3/3')

	def token_addrimport_addr2(self):
		return self._token_addrimport('token_addr2', '21-23', expect='3/3')

	def token_addrimport_batch(self):
		return self._token_addrimport('token_addr1', '11-13', expect='3 addresses', extra_args=['--batch'])

	def token_addrimport_sym(self):
		return self._addrimport(
			ext      = '[21-23]{}.regtest.addrs',
			expect   = '3/3',
			add_args = ['--token=MM2'])

	def bal7(self):
		return self.bal5()
	def token_bal1(self):
		return self.token_bal(n='1')

	def token_txcreate(
			self,
			args      = [],
			token     = '',
			inputs    = '1',
			fee       = '50G',
			file_desc = 'Unsigned transaction'):
		return self.txcreate_ui_common(
			self.spawn(
				'mmgen-txcreate',
				self.eth_opts + [f'--token={token}', '-B', f'--fee={fee}'] + args),
			menu              = [],
			inputs            = inputs,
			input_sels_prompt = 'to spend from',
			add_comment       = tx_comment_lat_cyr_gr,
			file_desc         = file_desc)
	def token_txsign(self, ext='', token='', add_args=[], ni=True):
		return self.txsign(ni=ni, ext=ext, add_args=add_args)
	def token_txsend(self, ext='', token=''):
		return self.txsend(ext=ext)

	def token_txcreate1(self):
		return self.token_txcreate(args=['98831F3A:E:12,1.23456'], token='mm1')
	def token_txview1_raw(self):
		return self.txview(ext_fs='1.23456,50000]{}.regtest.rawtx')
	def token_txsign1(self):
		return self.token_txsign(
			ext = '1.23456,50000]{}.regtest.rawtx',
			token = 'mm1',
			ni = False,
			add_args = ['--no-quiet', '--no-yes'])
	def token_txsend1(self):
		return self.token_txsend(ext='1.23456,50000]{}.regtest.sigtx', token='mm1')
	def token_txview1_sig(self):
		return self.txview(ext_fs='1.23456,50000]{}.regtest.sigtx')

	def tx_status3(self):
		return self.tx_status(
			ext         = '1.23456,50000]{}.regtest.sigtx',
			expect_str  = 'successfully executed',
			expect_str2 = 'has 1 confirmation')

	def token_bal2(self):
		return self.token_bal(n='2')

	def twview(self, args=[], expect_str='', tool_args=[]):
		t = self.spawn('mmgen-tool', self.eth_opts + args + ['twview'] + tool_args)
		if expect_str:
			t.expect(expect_str, regex=True)
		return t

	def token_txcreate2(self):
		return self.token_txcreate(args=[burn_addr+', '+amt2], token='mm1')
	def token_txbump(self):
		return self._txbump(ext=amt2+',50000]{}.regtest.rawtx', fee='56G', add_opts=['--token=MM1'])
	def token_txsign2(self):
		return self.token_txsign(ext=amt2+',50000]{}.regtest.rawtx', token='mm1')
	def token_txsend2(self):
		return self.token_txsend(ext=amt2+',50000]{}.regtest.sigtx', token='mm1')

	def token_bal3(self):
		return self.token_bal(n='3')

	def del_devaddr(self):
		return self._del_addr(dfl_devaddr)

	def bal1_getbalance(self):
		return self.bal_getbalance(dfl_sid, '1', etc_adj=True)

	def addrimport_token_burn_addr(self):
		return self._addrimport_one_addr(addr=burn_addr, extra_args=['--token=mm1'])

	def token_bal4(self):
		return self.token_bal(n='4')

	def token_bal_getbalance(self):
		return self.bal_getbalance(dfl_sid, '2', extra_args=['--token=mm1'])

	def txcreate_noamt(self):
		return self.txcreate(args=['98831F3A:E:12'])
	def txsign_noamt(self):
		return self.txsign(ext='99.99895,50000]{}.regtest.rawtx')
	def txsend_noamt(self):
		return self.txsend(ext='99.99895,50000]{}.regtest.sigtx')

	def bal8(self):
		return self.bal(n='8')
	def token_bal5(self):
		return self.token_bal(n='5')

	def token_txcreate_noamt(self):
		return self.token_txcreate(args=['98831F3A:E:13'], token='mm1', inputs='2', fee='51G')
	def token_txsign_noamt(self):
		return self.token_txsign(ext='1.23456,51000]{}.regtest.rawtx', token='mm1')
	def token_txsend_noamt(self):
		return self.token_txsend(ext='1.23456,51000]{}.regtest.sigtx', token='mm1')

	def bal9(self):
		return self.bal(n='9')
	def token_bal6(self):
		return self.token_bal(n='6')

	def listaddresses(self, args=[], tool_args=['all_labels=1']):
		return self.spawn('mmgen-tool', self.eth_opts + args + ['listaddresses'] + tool_args)

	def listaddresses1(self):
		return self.listaddresses()
	def listaddresses2(self):
		return self.listaddresses(tool_args=['minconf=3'])
	def listaddresses3(self):
		return self.listaddresses(tool_args=['sort=amt', 'reverse=1'])
	def listaddresses4(self):
		return self.listaddresses(tool_args=['sort=age', 'showempty=0'])

	def token_listaddresses1(self):
		return self.listaddresses(args=['--token=mm1'])
	def token_listaddresses2(self):
		return self.listaddresses(args=['--token=mm1'], tool_args=['showempty=1'])
	def token_listaddresses3(self):
		return self.listaddresses(args=['--token=mm1'], tool_args=['showempty=0'])
	def token_listaddresses4(self):
		return self.listaddresses(args=['--token=mm2'], tool_args=['sort=age', 'reverse=1'])

	def twview_cached_balances(self):
		return self.twview(args=['--cached-balances'])
	def token_twview_cached_balances(self):
		return self.twview(args=['--token=mm1', '--cached-balances'])

	def txcreate_cached_balances(self):
		args = ['--fee=20G', '--cached-balances', '98831F3A:E:3, 0.1276']
		return self.txcreate(args=args, acct='2')
	def token_txcreate_cached_balances(self):
		args=['--cached-balances', '--fee=12G', '98831F3A:E:12, 1.2789']
		return self.token_txcreate(args=args, token='mm1')

	def txdo_cached_balances(
			self,
			acct          = '2',
			fee_info_data = ('0.00105', '50'),
			add_args      = ['98831F3A:E:3,0.4321']):
		t = self.txcreate(
			args          = ['--fee=20G', '--cached-balances'] + add_args + [dfl_words_file],
			acct          = acct,
			caller        = 'txdo',
			fee_info_data = fee_info_data,
			no_read       = True)
		self._do_confirm_send(t, quiet=not self.cfg.debug, sure=False)
		return t

	def txcreate_refresh_balances(self):
		return self._txcreate_refresh_balances(
			bals       = ['2', '3'],
			args       = ['-B', '--cached-balances', '-i'],
			total      = vbal5,
			adj_total  = True,
			total_coin = None)

	def _txcreate_refresh_balances(self, bals, args, total, adj_total, total_coin):

		self.mining_delay()

		if total_coin is None:
			total_coin = self.proto.coin

		t = self.spawn('mmgen-txcreate', self.eth_opts + args)
		for n in bals:
			t.expect('[R]efresh balance:\b', 'R')
			t.expect(' main menu): ', n+'\n')
			t.expect('OK? (y/N): ', 'y')
		t.expect('[R]efresh balance:\b', 'q')
		t.expect(rf'Total unspent:.*\D{total}\D.*{total_coin}', regex=True)
		return t

	def bal10(self):
		return self.bal(n='10')

	def token_txdo_cached_balances(self):
		return self.txdo_cached_balances(
			acct          = '1',
			fee_info_data = (0.0025786 if self.daemon.id == 'parity' else '0.00260265', '50'),
			add_args      = ['--token=mm1', '98831F3A:E:12,43.21'])

	def token_txcreate_refresh_balances(self):
		return self._txcreate_refresh_balances(
			bals       = ['1', '2'],
			args       = ['--token=mm1', '-B', '--cached-balances', '-i'],
			total      = '1000',
			adj_total  = False,
			total_coin = 'MM1')

	def token_bal7(self):
		return self.token_bal(n='7')

	def twview1(self):
		return self.twview()
	def twview2(self):
		return self.twview(tool_args=['wide=1'])
	def twview3(self):
		return self.twview(tool_args=['wide=1', 'sort=age'])
	def twview4(self):
		return self.twview(tool_args=['wide=1', 'minconf=15'], expect_str=r'E:1\D.*\D100\D')
	def twview5(self):
		return self.twview(tool_args=['wide=1', 'minconf=0'])
	def twview6(self):
		return self.twview(expect_str=vbal7)
	def twview7(self):
		return self.twview(args=['--cached-balances'])
	def twview8(self):
		return self.twview()
	def twview9(self):
		return self.twview(args=['--cached-balances'], expect_str=vbal6)

	def token_twview1(self):
		return self.twview(args=['--token=mm1'])
	def token_twview2(self):
		return self.twview(args=['--token=mm1'], tool_args=['wide=1'])
	def token_twview3(self):
		return self.twview(args=['--token=mm1'], tool_args=['wide=1', 'sort=age'])

	def edit_comment(
			self,
			out_num,
			args          = [],
			action        = 'l',
			comment_text  = None,
			changed       = False,
			pexpect_spawn = None):

		t = self.spawn('mmgen-txcreate', self.eth_opts + args + ['-B', '-i'], pexpect_spawn=pexpect_spawn)

		t.expect(self.menu_prompt, 'M')
		t.expect(self.menu_prompt, action)
		t.expect(r'return to main menu): ', out_num+'\n')

		for p, r in (
			('Enter label text.*: ', comment_text+'\n') if comment_text is not None else (r'\(y/N\): ', 'y'),
			(r'\(y/N\): ', 'y') if comment_text == Ctrl_U else (None, None),
		):
			if p:
				t.expect(p, r, regex=True)

		m = (
			'Label for account #{} edited' if changed else
			'Account #{} removed' if action == 'D' else
			'Label added to account #{}' if comment_text and comment_text != Ctrl_U else
			'Label removed from account #{}')

		t.expect(m.format(out_num))
		t.expect(self.menu_prompt, 'M')
		t.expect(self.menu_prompt, 'q')

		t.expect('Total unspent:')

		return t

	def edit_comment1(self):
		return self.edit_comment(out_num=del_addrs[0], comment_text=tw_comment_zh[:3])
	def edit_comment2(self):
		spawn = not sys.platform == 'win32'
		return self.edit_comment(
			out_num       = del_addrs[0],
			comment_text  = tw_comment_zh[3:],
			changed       = True,
			pexpect_spawn = spawn)
	def edit_comment3(self):
		return self.edit_comment(out_num=del_addrs[1], comment_text=tw_comment_lat_cyr_gr)
	def edit_comment4(self):
		if self.skip_for_win('no pexpect_spawn'):
			return 'skip'
		return self.edit_comment(out_num=del_addrs[0], comment_text=Ctrl_U, pexpect_spawn=True)

	def token_edit_comment1(self):
		return self.edit_comment(out_num='1', comment_text='Token label #1', args=['--token=mm1'])

	def remove_addr1(self):
		return self.edit_comment(out_num=del_addrs[0], action='D')
	def remove_addr2(self):
		return self.edit_comment(out_num=del_addrs[1], action='D')
	def token_remove_addr1(self):
		return self.edit_comment(out_num=del_addrs[0], args=['--token=mm1'], action='D')
	def token_remove_addr2(self):
		return self.edit_comment(out_num=del_addrs[1], args=['--token=mm1'], action='D')

	def twexport_noamt(self):
		return self.twexport(add_args=['include_amts=0'])

	def twexport(self, add_args=[]):
		t = self.spawn('mmgen-tool', self.eth_opts + ['twexport'] + add_args)
		t.written_to_file('JSON data')
		return t

	async def twmove(self):
		self.spawn(msg_only=True)
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(self.cfg, self.proto, no_wallet_init=True)
		imsg('Moving tracking wallet')
		fn_bak = twctl.tw_path.with_suffix('.bak.json')
		fn_bak.unlink(missing_ok=True)
		twctl.tw_path.rename(fn_bak)
		return 'ok'

	def twimport(self, add_args=[], expect_str=None):
		from mmgen.tw.json import TwJSON
		fn = joinpath(self.tmpdir, TwJSON.Base(self.cfg, self.proto).dump_fn)
		t = self.spawn('mmgen-tool', self.eth_opts_noquiet + ['twimport', fn] + add_args)
		t.expect('(y/N): ', 'y')
		if expect_str:
			t.expect(expect_str)
		t.written_to_file('tracking wallet data')
		return t

	def twimport_nochksum(self):
		return self.twimport(add_args=['ignore_checksum=true'], expect_str='ignoring incorrect checksum')

	def tw_chktotal(self):
		self.spawn(msg_only=True)
		from mmgen.tw.json import TwJSON
		fn = joinpath(self.tmpdir, TwJSON.Base(self.cfg, self.proto).dump_fn)
		res = json.loads(read_from_file(fn))
		cmp_or_die(res['data']['value'], vbal6, 'value in tracking wallet JSON dump')
		return 'ok'

	async def twcompare(self):
		self.spawn(msg_only=True)
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(self.cfg, self.proto, no_wallet_init=True)
		fn = twctl.tw_path
		fn_bak = fn.with_suffix('.bak.json')
		imsg('Comparing imported tracking wallet with original')
		data = [json.dumps(json.loads(f.read_text()), sort_keys=True) for f in (fn, fn_bak)]
		cmp_or_die(*data, 'tracking wallets')
		return 'ok'

	def edit_json_twdump(self):
		self.spawn(msg_only=True)
		from mmgen.tw.json import TwJSON
		fn = TwJSON.Base(self.cfg, self.proto).dump_fn
		text = json.loads(self.read_from_tmpfile(fn))
		token_addr = self.read_from_tmpfile('token_addr2').strip()
		text['data']['entries']['tokens'][token_addr][2][3] = f'edited comment [фубар] [{gr_uc}]'
		self.write_to_tmpfile(fn, json.dumps(text, indent=4))
		return 'ok'

	def stop(self):
		self.spawn(msg_only=True)
		if self.cfg.no_daemon_stop:
			msg_r(f'(leaving {self.daemon.id} daemon running by user request)')
			imsg('')
		elif not stop_test_daemons(self.proto.coin+'_rt', remove_datadir=True):
			return False
		set_vt100()
		return 'ok'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_d.include.input: Shared input routines for the cmdtest.py test suite
"""

import time
from ...include.common import getrand
from .common import randbool

def stealth_mnemonic_entry(t, mne, mn, entry_mode, pad_entry=False):

	def pad_mnemonic(mn, ss_len):
		def get_pad_chars(n):
			ret = ''
			for _ in range(n):
				m = int.from_bytes(getrand(1), 'big') % 32
				ret += r'123579!@#$%^&*()_+-=[]{}"?/,.<>|'[m]
			return ret
		ret = []
		for w in mn:
			if entry_mode == 'short':
				w = w[:ss_len]
				if len(w) < ss_len:
					npc = 3
					w = w[0] + get_pad_chars(npc) + w[1:]
					if pad_entry:
						w += '%' * (1 + mne.em.pad_max - npc)
					else:
						w += '\n'
				else:
					w = get_pad_chars(1) + w[0] + get_pad_chars(1) + w[1:]
			elif len(w) > (3, 5)[ss_len==12]:
				w = w + '\n'
			else:
				w = (
					get_pad_chars(2 if randbool() and entry_mode != 'short' else 0)
					+ w[0] + get_pad_chars(2) + w[1:]
					+ get_pad_chars(9))
				w = w[:ss_len+1]
			ret.append(w)
		return ret

	if entry_mode == 'fixed':
		mn = ['bkr'] + mn[:5] + ['nfb'] + mn[5:]
		ssl = mne.uniq_ss_len
		def gen_mn():
			for w in mn:
				if len(w) >= ssl:
					yield w[:ssl]
				else:
					yield w[0] + 'z\b' + '#' * (ssl-len(w)) + w[1:]
		mn = list(gen_mn())
	elif entry_mode in ('full', 'short'):
		mn = ['fzr'] + mn[:5] + ['grd', 'grdbxm'] + mn[5:]
		mn = pad_mnemonic(mn, mne.em.ss_len)
		mn[10] = '@#$%*##' + mn[10]

	wnum = 1
	p_ok, p_err = mne.word_prompt
	for w in mn:
		ret = t.expect((p_ok.format(wnum), p_err.format(wnum-1)))
		if ret == 0:
			wnum += 1
		for char in w:
			t.send(char)
			time.sleep(0.005)

def user_dieroll_entry(t, data):
	for s in data:
		t.expect(r'Enter die roll #.+: ', s, regex=True)
		time.sleep(0.005)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.online: THORChain online signed transaction class
"""

from ....util import msg, pp_msg, die
from ....tx import online as TxBase

from .signed import Signed

class OnlineSigned(Signed, TxBase.OnlineSigned):

	async def test_sendable(self, txhex):
		res = self.rpc.tx_op(txhex, op='check_tx')
		if res['code'] == 0:
			return True
		else:
			pp_msg(res)
			return False

	async def send_checks(self):
		pass

	async def send_with_node(self, txhex):
		res = self.rpc.tx_op(txhex, op='broadcast_tx_sync') # broadcast_tx_async
		if res['code'] == 0:
			return res['hash'].lower()
		else:
			pp_msg(res)
			die(2, 'Transaction send failed')

	async def post_network_send(self, coin_txid):
		return True

	async def get_receipt(self, txid, *, receipt_only=False):
		try:
			return self.rpc.get_tx_info(txid)
		except Exception as e:
			msg(f'{type(e).__name__}: {e}')
			return False

class Sent(TxBase.Sent, OnlineSigned):
	pass

class AutomountOnlineSigned(TxBase.AutomountOnlineSigned, OnlineSigned):
	pass

class AutomountSent(TxBase.AutomountSent, AutomountOnlineSigned):
	pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.mnemonic: Mnemonic routines for the 'mmgen-tool' utility
"""

from collections import namedtuple

from .common import tool_cmd_base, options_annot_str

from ..baseconv import baseconv
from ..xmrseed import xmrseed
from ..bip39 import bip39

dfl_mnemonic_fmt = 'mmgen'
mft = namedtuple('mnemonic_format', ['fmt', 'pad', 'conv_cls'])
mnemonic_fmts = {
	'mmgen':   mft('words',  'seed', baseconv),
	'bip39':   mft('bip39',   None,  bip39),
	'xmrseed': mft('xmrseed', None,  xmrseed),
}
mn_opts_disp = 'seed phrase format ' + options_annot_str(mnemonic_fmts)

class tool_cmd(tool_cmd_base):
	"""
	seed phrase utilities

	Supported seed phrase formats: 'mmgen' (default), 'bip39', 'xmrseed'

	IMPORTANT NOTE: MMGen Wallet’s default seed phrase format uses the
	Electrum wordlist, however seed phrases are computed using a different
	algorithm and are NOT Electrum-compatible!

	BIP39 support is fully compatible with the standard, allowing users to
	import and export seed entropy from BIP39-compatible wallets.  However,
	users should be aware that BIP39 support does not imply BIP32 support!
	MMGen uses its own key derivation scheme differing from the one described
	by the BIP32 protocol.

	For Monero (‘xmrseed’) seed phrases, input data is reduced to a spendkey
	before conversion so that a canonical seed phrase is produced.  This is
	required because Monero seeds, unlike ordinary wallet seeds, are tied
	to a concrete key/address pair.  To manually generate a Monero spendkey,
	use the ‘hex2wif’ command.
	"""

	def _xmr_reduce(self, bytestr):
		from ..protocol import init_proto
		proto = init_proto(self.cfg, 'xmr')
		if len(bytestr) != proto.privkey_len:
			from ..util import die
			die(1, '{!r}: invalid bit length for Monero private key (must be {})'.format(
				len(bytestr*8),
				proto.privkey_len*8))
		return proto.preprocess_key(bytestr, None)

	def _do_random_mn(self, nbytes: int, fmt: str):
		assert nbytes in (16, 24, 32), 'nbytes must be 16, 24 or 32'
		from ..crypto import Crypto
		randbytes = Crypto(self.cfg).get_random(nbytes)
		if fmt == 'xmrseed':
			randbytes = self._xmr_reduce(randbytes)
		if self.cfg.verbose:
			from ..util import msg
			msg(f'Seed: {randbytes.hex()}')
		return self.hex2mn(randbytes.hex(), fmt=fmt)

	def mn_rand128(self, fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"generate a random 128-bit mnemonic seed phrase"
		return self._do_random_mn(16, fmt)

	def mn_rand192(self, fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"generate a random 192-bit mnemonic seed phrase"
		return self._do_random_mn(24, fmt)

	def mn_rand256(self, fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"generate a random 256-bit mnemonic seed phrase"
		return self._do_random_mn(32, fmt)

	def hex2mn(self, hexstr: 'sstr', fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"convert a 16, 24 or 32-byte hexadecimal string to a mnemonic seed phrase"
		if fmt == 'xmrseed':
			hexstr = self._xmr_reduce(bytes.fromhex(hexstr)).hex()
		f = mnemonic_fmts[fmt]
		return ' '.join(f.conv_cls(fmt).fromhex(hexstr, pad=f.pad))

	def mn2hex(self, seed_mnemonic: 'sstr', fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"convert a mnemonic seed phrase to a hexadecimal string"
		f = mnemonic_fmts[fmt]
		return f.conv_cls(fmt).tohex(seed_mnemonic.split(), pad=f.pad)

	def mn2hex_interactive(self,
			*,
			fmt: mn_opts_disp = dfl_mnemonic_fmt,
			mn_len: 'length of seed phrase in words' = 24,
			print_mn: 'print the seed phrase after entry' = False):
		"convert an interactively supplied mnemonic seed phrase to a hexadecimal string"
		from ..mn_entry import mn_entry
		mn = mn_entry(self.cfg, fmt).get_mnemonic_from_user(25 if fmt == 'xmrseed' else mn_len, validate=False)
		if print_mn:
			from ..util import msg
			msg(mn)
		return self.mn2hex(seed_mnemonic=mn, fmt=fmt)

	def mn_stats(self, fmt:mn_opts_disp = dfl_mnemonic_fmt):
		"show stats for a mnemonic wordlist"
		return mnemonic_fmts[fmt].conv_cls(fmt).check_wordlist(self.cfg)

	def mn_printlist(self,
			*,
			fmt: mn_opts_disp = dfl_mnemonic_fmt,
			enum: 'enumerate the list' = False,
			pager: 'send output to pager' = False):
		"print a mnemonic wordlist"
		ret = mnemonic_fmts[fmt].conv_cls(fmt).get_wordlist()
		if enum:
			ret = [f'{n:>4} {e}' for n, e in enumerate(ret)]
		return '\n'.join(ret)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.protobuf: transaction serialization for THORChain protocol
"""

# https://dev.thorchain.org/technical-faq.html

from decimal import Decimal
from collections import namedtuple
from dataclasses import dataclass #, field
from typing import Annotated, List, Optional
from pure_protobuf.annotations import Field
from pure_protobuf.message import BaseMessage

from ...secp256k1.util import sign_message

from ...cosmos.tx.protobuf import (
	ModeInfo,
	SignDoc,
	SignerInfo,
	SignMode,
	PublicKey,
	PublicKeyData,
	Coin,
	Fee,
	AuthInfo,
	Tx,
	TxMsg)

send_tx_parms = namedtuple(
	'rune_send_tx_parms', [
		'from_addr',
		'to_addr',
		'amt',
		'gas_limit',
		'account_number',
		'sequence',
		'pubkey',
		'wifkey',
		'signature'],
		defaults = (None, None, None))

deposit_tx_parms = namedtuple(
	'rune_deposit_tx_parms', [
		'chain',
		'symbol',
		'ticker',
		'from_addr',
		'amt',
		'gas_limit',
		'account_number',
		'sequence',
		'decimals',
		'memo',
		'synth', # begin default fields
		'trade',
		'secured',
		'pubkey',
		'wifkey',
		'signature'],
		defaults = (None, None, None, None, None, None))

# subset of deposit_tx_parms:
swap_tx_parms = namedtuple(
	'rune_swap_tx_parms', [
		'from_addr',
		'amt',
		'gas_limit',
		'account_number',
		'sequence',
		'memo',
		'pubkey', # begin default fields
		'wifkey',
		'signature'],
		defaults = (None, None, None))

@dataclass
class Asset(BaseMessage):
	chain:   Annotated[str, Field(1)]
	symbol:  Annotated[str, Field(2)]
	ticker:  Annotated[str, Field(3)]
	synth:   Annotated[Optional[bool], Field(4)] = None
	trade:   Annotated[Optional[bool], Field(5)] = None
	secured: Annotated[Optional[bool], Field(6)] = None

@dataclass
class CoinWithAsset(BaseMessage):
	asset:    Annotated[Asset, Field(1)]
	amount:   Annotated[str, Field(2)]
	decimals: Annotated[Optional[int], Field(3)] = None

class Messages:

	@dataclass
	class MsgSend(BaseMessage):

		@dataclass
		class Body(BaseMessage):
			fromAddress: Annotated[bytes, Field(1)]
			toAddress:   Annotated[bytes, Field(2)]
			amount:      Annotated[List[Coin], Field(3)]

		id:   Annotated[str, Field(1)] # '/types.MsgSend'
		body: Annotated[Body, Field(2)]

	@dataclass
	class MsgDeposit(BaseMessage):
	# To initiate a $RUNE -> $ASSET swap a MsgDeposit must be broadcasted to the THORChain blockchain.
	# The MsgDeposit does not have a destination address, and has the following properties.
	# MsgDeposit{
	#     Coins:  coins,
	#     Memo:   memo,
	#     Signer: signer,
	# }
		@dataclass
		class Body(BaseMessage):
			coins:  Annotated[List[CoinWithAsset], Field(1)]
			memo:   Annotated[str, Field(2)]
			signer: Annotated[bytes, Field(3)]

		id:   Annotated[str, Field(1)] # '/types.MsgDeposit'
		body: Annotated[Body, Field(2)]

@dataclass
class RuneTxMsg(TxMsg):
	msgs_cls = Messages
	id:        Annotated[str, Field(1)]
	bodyBytes: Annotated[bytes, Field(2)]

@dataclass
class RuneTxBody(BaseMessage):
	messages:                    Annotated[List[RuneTxMsg], Field(1)]
	memo:                        Annotated[Optional[str], Field(2)] = None
	timeoutHeight:               Annotated[Optional[int], Field(3)] = None
	extensionOptions:            Annotated[Optional[bytes], Field(4)] = None
	nonCriticalExtensionOptions: Annotated[Optional[bytes], Field(5)] = None

@dataclass
class RuneTx(Tx):
	body:       Annotated[RuneTxBody, Field(1)]
	authInfo:   Annotated[AuthInfo, Field(2)]
	signatures: Annotated[List[bytes], Field(3)]

def amt_to_base_unit(amt, *, decimals):
	return int(Decimal(amt) * (Decimal('10') ** decimals))

def base_unit_to_amt(n, *, decimals):
	return n * Decimal('10') ** -decimals

def tx_info(tx, proto):
	b = tx.body.messages[0].body
	s = tx.authInfo.signerInfos[0]
	msg_type = tx.body.messages[0].id.removeprefix('/types.')
	if msg_type == 'MsgSend':
		from_addr = proto.encode_addr_bech32x(b.fromAddress)
		to_addr   = proto.encode_addr_bech32x(b.toAddress)
		asset     = b.amount[0].denom.upper()
		memo      = tx.body.memo
		amt       = base_unit_to_amt(int(b.amount[0].amount), decimals=8)
	elif msg_type == 'MsgDeposit':
		from_addr = proto.encode_addr_bech32x(b.signer)
		to_addr = 'None'
		asset     = b.coins[0].asset.symbol
		memo      = b.memo
		amt       = base_unit_to_amt(int(b.coins[0].amount), decimals=b.coins[0].decimals or 8)
	yield f'TxID:      {tx.txid}'
	yield f'Type:      {msg_type}'
	yield f'From:      {from_addr}'
	yield f'To:        {to_addr}'
	yield f'Asset:     {asset}'
	yield f'Amount:    {amt}'
	yield f'Sequence:  {int(s.sequence)}'
	yield f'Gas limit: {tx.authInfo.fee.gasLimit}'
	yield f'Memo:      {memo}'
	yield f'Pubkey:    {s.publicKey.key.data.hex()}'

def build_swap_tx(cfg, proto, parms, *, skip_body_memo=False):

	p = parms
	assert type(p) == swap_tx_parms, f'{p}: invalid ‘parms’ (not swap_tx_parms instance)'

	return build_tx(
			cfg,
			proto,
			deposit_tx_parms(
				chain = 'THOR',
				symbol = 'RUNE',
				ticker = 'RUNE',
				from_addr = p.from_addr,
				amt = p.amt,
				gas_limit = p.gas_limit,
				account_number = p.account_number,
				sequence = p.sequence,
				decimals = 8,
				memo = p.memo,
				pubkey = p.pubkey,
				wifkey = p.wifkey,
				signature = p.signature),
			skip_body_memo = skip_body_memo)

def build_tx(cfg, proto, parms, *, null_fee=False, skip_body_memo=False):

	p = parms
	assert type(p) in (send_tx_parms, deposit_tx_parms), f'{p}: invalid ‘parms’ (not *_tx_parms instance)'

	msg_type = 'MsgSend' if type(p) == send_tx_parms else 'MsgDeposit'

	if p.wifkey:
		assert p.pubkey is None and p.signature is None
		from ....key import PrivKey
		from ....keygen import KeyGenerator
		privkey = PrivKey(proto=proto, wif=p.wifkey)
		pubkey = KeyGenerator(cfg, proto, 'std').to_pubkey(privkey)
	else:
		assert p.pubkey and p.signature
		pubkey = p.pubkey

	cls = getattr(Messages, msg_type)

	if msg_type == 'MsgSend':
		message = cls(
			id   = '/types.MsgSend',
			body = cls.Body(
				fromAddress = proto.decode_addr(p.from_addr).bytes,
				toAddress = proto.decode_addr(p.to_addr).bytes,
				amount = [Coin(denom='rune', amount=str(amt_to_base_unit(p.amt, decimals=8)))]))
		fee_amt = None
	elif msg_type == 'MsgDeposit':
		coin_data = CoinWithAsset(
			asset = Asset(
				chain = p.chain,
				symbol = p.symbol,
				ticker = p.ticker,
				synth = p.synth,
				trade = p.trade,
				secured = p.secured),
			amount = str(amt_to_base_unit(p.amt, decimals=p.decimals or 8)),
			decimals = p.decimals)
		message = cls(
			id   = '/types.MsgDeposit',
			body = cls.Body(
				coins = [coin_data],
				memo = p.memo,
				signer = proto.decode_addr(p.from_addr).bytes))
		fee_amt = None if null_fee else [Coin(denom='rune', amount='0')]

	signer_info = SignerInfo(
		publicKey = PublicKey(
			id = '/cosmos.crypto.secp256k1.PubKey',
			key = PublicKeyData(data=pubkey)),
		modeInfo = ModeInfo(
			single = ModeInfo.Single(mode=SignMode.SIGN_MODE_DIRECT),
			multi = None),
		sequence = p.sequence)

	body = RuneTxBody(messages=[message], memo=None if skip_body_memo else getattr(p, 'memo', None))

	auth_info = AuthInfo(signerInfos=[signer_info], fee=Fee(gasLimit=p.gas_limit, amount=fee_amt))

	# cosmjs/packages/crypto/src/secp256k1signature.ts
	# cosmjs/packages/amino/src/signature.ts
	#   Signature must be 64 bytes long. Cosmos SDK uses a 2x32 byte fixed length
	#   encoding for the secp256k1 signature integers r and s.
	signature = sign_message(
		sign_doc = SignDoc(
			bodyBytes = bytes(body),
			authInfoBytes = bytes(auth_info),
			chainId = proto.chain_id,
			accountNumber = p.account_number),
		sec_bytes = privkey) if p.wifkey else p.signature

	return RuneTx(body=body, authInfo=auth_info, signatures=[signature])



#!/usr/bin/env python3

"""
test.modtest_d.addrlist: address list unit tests for the MMGen suite
"""

from mmgen.color import blue
from mmgen.util import msg

from mmgen.seed import Seed
from mmgen.addr import MMGenAddrType
from mmgen.addrlist import AddrIdxList, AddrList, KeyList, KeyAddrList, ViewKeyAddrList
from mmgen.passwdlist import PasswordList
from mmgen.protocol import init_proto
from ..include.common import cfg, qmsg, vmsg

def do_test(
		list_type,
		chksum,
		idx_spec   = None,
		pw_id_str  = None,
		add_kwargs = None,
		coin       = None,
		addrtype   = None):

	qmsg(blue(f'Testing {list_type.__name__}'))
	proto = init_proto(cfg, coin or 'btc')
	seed = Seed(cfg, seed_bin=bytes.fromhex('feedbead'*8))
	mmtype = MMGenAddrType(proto, addrtype or 'C')
	idxs = AddrIdxList(fmt_str=idx_spec or '1-3')

	if cfg.verbose:
		debug_addrlist_save = cfg.debug_addrlist
		cfg.debug_addrlist = True

	kwargs = {
		'seed': seed,
		'pw_idxs': idxs,
		'pw_id_str': pw_id_str,
		'pw_fmt': 'b58',
	} if pw_id_str else {
		'seed': seed,
		'addr_idxs': idxs,
		'mmtype': mmtype,
	}

	if add_kwargs:
		kwargs.update(add_kwargs)

	al = list_type(cfg, proto, **kwargs)

	al.file.format()

	qmsg(f'Filename: {al.file.filename}\n')

	vmsg(f'------------\n{al.file.fmt_data}\n------------')

	if chksum:
		assert al.chksum == chksum, f'{al.chksum} != {chksum}'

	if cfg.verbose:
		cfg.debug_addrlist = debug_addrlist_save

	return True

class unit_tests:

	altcoin_deps = ('keyaddr_xmr', 'viewkeyaddr')

	def idxlist(self, name, ut):
		for i, o in (
				('99,88-102,1-3,4,9,818,444-445,816',        '1-4,9,88-102,444-445,816,818'),
				('99,88-99,100,102,4-7,9,818,444-445,816,1', '1,4-7,9,88-100,102,444-445,816,818'),
				('8',             '8'),
				('2-4',           '2-4'),
				('1,2-4',         '1-4'),
				('2-4,1-9,9,1,8', '1-9'),
				('2-4,1',         '1-4'),
				('2-2',           '2'),
				('2,2',           '2'),
				('2-3',           '2-3'),
				('2,3',           '2-3'),
				('3,2',           '2-3'),
				('2,4',           '2,4'),
				('',              ''),
			):
			l = AddrIdxList(fmt_str=i)
			if cfg.verbose:
				msg(f'list: {list(l)}\nin:   {i}\nout:  {o}\n')
			assert l.id_str == o, f'{l.id_str} != {o}'

		return True

	def addr(self, name, ut):
		return (
			do_test(AddrList, 'BCE8 082C 0973 A525', '1-3') and
			do_test(AddrList, '88FA B04B A380 C1CB', '199999,99-101,77-78,7,3,2-9')
		)

	def key(self, name, ut):
		return do_test(KeyList, None)

	def keyaddr(self, name, ut):
		return do_test(KeyAddrList, '4A36 AA65 8C2B 7C35')

	def keyaddr_xmr(self, name, ut):
		return do_test(KeyAddrList, 'AAA2 BA69 17FC 9A88', coin='XMR', addrtype='M')

	def viewkeyaddr(self, name, ut):
		return do_test(ViewKeyAddrList, 'C122 2E58 DC28 D6AE', coin='XMR', addrtype='M')

	def passwd(self, name, ut):
		return do_test(PasswordList, 'FF4A B716 4513 8F8F', pw_id_str='foo')

	def passwd_bip39(self, name, ut):
		return do_test(PasswordList, 'C3A8 B2B2 1AA1 FB40', pw_id_str='foo', add_kwargs={'pw_fmt': 'bip39'})



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
msg: base message signing classes
"""

import os, importlib, json
from .cfg import gc
from .objmethods import MMGenObject, HiliteStr, InitErrors
from .util import msg, die, make_chksum_6, fmt_list, remove_dups
from .color import red, orange, grnbg
from .protocol import init_proto
from .fileutil import get_data_from_file, write_data_to_file
from .addr import MMGenID, CoinAddr

class MMGenIDRange(HiliteStr, InitErrors, MMGenObject):
	"""
	closely based on MMGenID
	"""
	color = 'orange'
	width = 0
	trunc_ok = False
	def __new__(cls, proto, id_str):
		from .addrlist import AddrIdxList
		from .addr import AddrListID
		from .seed import SeedID
		try:
			ss = str(id_str).split(':')
			assert len(ss) in (2, 3), 'not 2 or 3 colon-separated items'
			t = proto.addr_type((ss[1], proto.dfl_mmtype)[len(ss)==2])
			me = str.__new__(cls, '{}:{}:{}'.format(ss[0], t, ss[-1]))
			me.sid = SeedID(sid=ss[0])
			me.idxlist = AddrIdxList(fmt_str=ss[-1])
			me.mmtype = t
			assert t in proto.mmtypes, f'{t}: invalid address type for {proto.cls_name}'
			me.al_id = str.__new__(AddrListID, me.sid+':'+me.mmtype) # checks already done
			me.proto = proto
			return me
		except Exception as e:
			return cls.init_fail(e, id_str)

class coin_msg:

	class base(MMGenObject):

		ext = 'rawmsg.json'
		signed = False
		chksum_keys = ('addrlists', 'message', 'msghash_type', 'network')

		@property
		def desc(self):
			return ('signed' if self.signed else 'unsigned') + ' message data'

		@property
		def chksum(self):
			return make_chksum_6(
				json.dumps(
					{k:self.data[k] for k in self.chksum_keys},
					sort_keys = True,
					separators = (',', ':')
			))

		@property
		def filename_stem(self):
			coin, network = self.data['network'].split('_')
			return '{}[{}]{}'.format(
				self.chksum.upper(),
				coin.upper(),
				('' if network == 'mainnet' else '.'+network))

		@property
		def filename(self):
			return f'{self.filename_stem}.{self.ext}'

		@property
		def signed_filename(self):
			return f'{self.filename_stem}.{coin_msg.signed.ext}'

		@staticmethod
		def get_proto_from_file(cfg, filename):
			data = json.loads(get_data_from_file(cfg, filename))
			network_id = data['metadata']['network'] if 'metadata' in data else data['network'].lower()
			coin, network = network_id.split('_')
			return init_proto(cfg=cfg, coin=coin, network=network)

		def write_to_file(self, *, outdir=None, ask_overwrite=False):
			data = {
				'id': f'{gc.proj_name} {self.desc}',
				'metadata': self.data,
				'signatures': self.sigs,
			}

			write_data_to_file(
				cfg           = self.cfg,
				outfile       = os.path.join(outdir or '', self.filename),
				data          = json.dumps(data, sort_keys=True, indent=4),
				desc          = self.desc,
				ask_overwrite = ask_overwrite)

	class new(base):

		def __init__(self, message, addrlists, msghash_type, *args, **kwargs):

			msghash_type = msghash_type or self.msg_cls.msghash_types[0]

			if msghash_type not in self.msg_cls.msghash_types:
				die(2, f'msghash_type {msghash_type!r} not supported for {self.proto.base_proto} protocol')

			self.data = {
				'network': '{}_{}'.format(self.proto.coin.lower(), self.proto.network),
				'addrlists': [MMGenIDRange(self.proto, i) for i in addrlists.split()],
				'message': message,
				'msghash_type': msghash_type,
			}
			self.sigs = {}

	class completed(base):

		def __init__(self, data, infile, *args, **kwargs):

			if data:
				self.__dict__ = data
				return

			self.data = get_data_from_file(
				cfg    = self.cfg,
				infile = infile,
				desc   = self.desc)

			d = json.loads(self.data)
			self.data = d['metadata']
			self.sigs = d['signatures']
			self.addrlists = [MMGenIDRange(self.proto, i) for i in self.data['addrlists']]

		def format(self, req_addr=None):

			labels = {
				'addr':       'address:',
				'addr_p2pkh': 'addr_p2pkh:',
				'pubhash':    'pubkey hash:',
				'sig':        'signature:',
			}

			def gen_entry(e):
				for k in labels:
					if e.get(k):
						yield fs_sig.format(labels[k], e[k])

			def gen_all():
				for k, v in hdr_data.items():
					yield fs_hdr.format(v[0], v[1](self.data[k]))
				if self.sigs:
					yield ''
					yield 'Signatures:'
					for n, (k, v) in enumerate(self.sigs.items()):
						yield ''
						yield f'{n+1:>3}) {k}'
						yield from gen_entry(v)

			def gen_single():
				for k, v in hdr_data.items():
					yield fs_hdr.format(v[0], v[1](self.data[k]))
				if self.sigs:
					yield 'Signature data:'
					k = (
						CoinAddr(self.proto, req_addr) if type(self).__name__ == 'exported_sigs' else
						MMGenID(self.proto, req_addr))
					if k not in self.sigs:
						die(1, f'{k}: address not found in signature data')
					yield from gen_entry(self.sigs[k])

			hdr_data = {
				'message':      ('Message:',           grnbg),
				'network':      ('Network:',           lambda v: v.replace('_', ' ').upper()),
				'msghash_type': ('Message Hash Type:', lambda v: v),
				'addrlists':    ('Address Ranges:',    lambda v: fmt_list(v, fmt='bare')),
				'failed_sids':  ('Failed Seed IDs:',   lambda v: red(fmt_list(v, fmt='bare'))),
			}

			if len(self.msg_cls.msghash_types) == 1:
				del hdr_data['msghash_type']

			if req_addr or type(self).__name__ == 'exported_sigs':
				del hdr_data['addrlists']

			if req_addr or not self.data.get('failed_sids'):
				del hdr_data['failed_sids']

			fs_hdr = '{:%s} {}' % max(len(v[0]) for v in hdr_data.values())
			fs_sig = '%s{:%s} %s{}' % (
				' ' * (2 if req_addr else 5),
				max(len(labels[k]) for v in self.sigs.values() for k in v.keys()),
				self.msg_cls.sigdata_pfx or ''
			) if self.sigs else None

			if req_addr:
				return '\n'.join(gen_single())
			else:
				return ('' if self.sigs else 'UN') + 'SIGNED MESSAGE DATA:\n\n  ' + '\n  '.join(gen_all())

	class unsigned(completed):

		async def sign(self, wallet_files, *, passwd_file=None):

			from .addrlist import KeyAddrList

			async def sign_list(al_in, seed):
				al = KeyAddrList(
					cfg         = self.cfg,
					proto       = self.proto,
					seed        = seed,
					addr_idxs   = al_in.idxlist,
					mmtype      = al_in.mmtype,
					skip_chksum = True,
					add_p2pkh   = al_in.mmtype in ('S', 'B'))

				for e in al.data:
					sig = await self.do_sign(
						wif     = e.sec.wif,
						message = self.data['message'],
						msghash_type = self.data['msghash_type'])

					mmid = f'{al_in.sid}:{al_in.mmtype}:{e.idx}'
					data = {
						'addr': e.addr,
						'sig': sig,
					}
					if self.msg_cls.include_pubhash:
						data.update({'pubhash': self.proto.decode_addr(e.addr_p2pkh or e.addr).bytes.hex()})

					if e.addr_p2pkh:
						data.update({'addr_p2pkh': e.addr_p2pkh})

					self.sigs[mmid] = data

			if self.proto.sign_mode == 'daemon':
				from .rpc import rpc_init
				self.rpc = await rpc_init(self.cfg, self.proto, ignore_wallet=True)

			from .wallet import Wallet
			wallet_seeds = [Wallet(cfg=self.cfg, fn=fn, passwd_file=passwd_file).seed for fn in wallet_files]
			need_sids = remove_dups([al.sid for al in self.addrlists], quiet=True)
			saved_seeds = []

			# First try wallet seeds:
			for sid in need_sids:
				for seed in wallet_seeds:
					if sid == seed.sid:
						saved_seeds.append(seed)
						need_sids.remove(sid)
						break

			# Then subseeds:
			for sid in need_sids:
				for seed in wallet_seeds:
					subseed = seed.subseeds.get_subseed_by_seed_id(sid, print_msg=True)
					if subseed:
						saved_seeds.append(subseed)
						need_sids.remove(sid)
						break

			for al in self.addrlists:
				for seed in saved_seeds:
					if al.sid == seed.sid:
						await sign_list(al, seed)
						break

			if need_sids:
				msg('Failed Seed IDs: {}'.format(orange(fmt_list(need_sids, fmt='bare'))))

			self.data['failed_sids'] = need_sids

	class signed(completed):

		ext = 'sigmsg.json'
		signed = True

	class signed_online(signed):

		def get_sigs(self, addr):

			if addr:
				req_addr = (
					CoinAddr(self.proto, addr) if type(self).__name__ == 'exported_sigs' else
					MMGenID(self.proto, addr))
				sigs = {k:v for k, v in self.sigs.items() if k == req_addr}
			else:
				sigs = self.sigs

			if not sigs:
				die(1, 'No signatures')

			return sigs

		async def verify(self, *, addr=None):

			sigs = self.get_sigs(addr)

			if self.proto.sign_mode == 'daemon':
				from .rpc import rpc_init
				self.rpc = await rpc_init(self.cfg, self.proto, ignore_wallet=True)

			for k, v in sigs.items():
				ret = await self.do_verify(
					addr    = v.get('addr_p2pkh') or v['addr'],
					sig     = v['sig'],
					message = self.data['message'],
					msghash_type = self.data['msghash_type'])
				if not ret:
					die(3, f'Invalid signature for address {k} ({v["addr"]})')

			return len(sigs)

		def get_json_for_export(self, *, addr=None):
			sigs = list(self.get_sigs(addr).values())
			pfx = self.msg_cls.sigdata_pfx
			if pfx:
				sigs = [{k:pfx+v for k, v in e.items()} for e in sigs]
			return json.dumps(
				{
					'message': self.data['message'],
					'msghash_type': self.data['msghash_type'],
					'network': self.data['network'].upper(),
					'signatures': sigs,
				},
				sort_keys = True,
				indent = 4
			)

	class exported_sigs(signed_online):

		def __init__(self, infile, *args, **kwargs):

			self.data = json.loads(
				get_data_from_file(
					cfg    = self.cfg,
					infile = infile,
					desc   = self.desc)
				)

			pfx = self.msg_cls.sigdata_pfx
			self.sigs = {sig_data['addr']:sig_data for sig_data in (
				[{k:v[len(pfx):] for k, v in e.items()} for e in self.data['signatures']]
					if pfx else
				self.data['signatures']
			)}

def _get_obj(clsname, cfg, *args, coin=None, network='mainnet', infile=None, data=None, **kwargs):

	assert not args, 'msg:_get_obj(): only keyword args allowed'

	if clsname == 'signed':
		assert data and not (coin or infile), 'msg:_get_obj(): chk2'
	else:
		assert not data and (coin or infile) and not (coin and infile), 'msg:_get_obj(): chk3'

	proto = (
		data['proto'] if data else
		init_proto(cfg=cfg, coin=coin, network=network) if coin else
		coin_msg.base.get_proto_from_file(cfg, infile))

	try:
		msg_cls = getattr(
			importlib.import_module(f'mmgen.proto.{proto.base_proto_coin.lower()}.msg'),
			'coin_msg')
	except:
		die(1, f'Message signing operations not supported for {proto.base_proto} protocol')

	me = MMGenObject.__new__(getattr(msg_cls, clsname, getattr(coin_msg, clsname)))
	me.msg_cls = msg_cls
	me.cfg = cfg
	me.proto = proto

	me.__init__(infile=infile, data=data, *args, **kwargs)

	return me

def _get(clsname):
	return lambda *args, **kwargs: _get_obj(clsname, *args, **kwargs)

NewMsg          = _get('new')
CompletedMsg    = _get('completed')
UnsignedMsg     = _get('unsigned')
SignedMsg       = _get('signed')
SignedOnlineMsg = _get('signed_online')
ExportedMsgSigs = _get('exported_sigs')



#!/usr/bin/env python3

"""
test.modtest_d.cashaddr: unit test for the BCH cashaddr module
"""

from collections import namedtuple

from mmgen.proto.bch.cashaddr import cashaddr_parse_addr, cashaddr_decode_addr, cashaddr_encode_addr
from mmgen.addr import CoinAddr

from ..include.common import cfg, vmsg

from mmgen.protocol import init_proto
proto = init_proto(cfg, 'bch')

# Source: https://upgradespecs.bitcoincashnode.org/cashaddr
alias_data = """
1BpEi6DfDAUFd7GtittLSdBeYJvcoaVggu  bitcoincash:qpm2qsznhks23z7629mms6s4cwef74vcwvy22gdx6a
1KXrWXciRDZUpQwQmuM1DbwsKDLYAYsVLR  bitcoincash:qr95sy3j9xwd2ap32xkykttr4cvcu7as4y0qverfuy
16w1D5WRVKJuZUsSRzdLp9w3YGcgoxDXb   bitcoincash:qqq3728yw0y47sqn6l2na30mcw6zm78dzqre909m2r
3CWFddi6m4ndiGyKqzYvsFYagqDLPVMTzC  bitcoincash:ppm2qsznhks23z7629mms6s4cwef74vcwvn0h829pq
3LDsS579y7sruadqu11beEJoTjdFiFCdX4  bitcoincash:pr95sy3j9xwd2ap32xkykttr4cvcu7as4yc93ky28e
31nwvkZwyPdgzjBJZXfDmSWsC4ZLKpYyUw  bitcoincash:pqq3728yw0y47sqn6l2na30mcw6zm78dzq5ucqzc37
"""

vec_data = """
F5BF48B397DAE70BE82B3CCA4793F8EB2B6CDAC9
20 0  bitcoincash:qr6m7j9njldwwzlg9v7v53unlr4jkmx6eylep8ekg2
20 1  bchtest:pr6m7j9njldwwzlg9v7v53unlr4jkmx6eyvwc0uz5t
20 1  pref:pr6m7j9njldwwzlg9v7v53unlr4jkmx6ey65nvtks5
20 15 prefix:0r6m7j9njldwwzlg9v7v53unlr4jkmx6ey3qnjwsrf

7ADBF6C17084BC86C1706827B41A56F5CA32865925E946EA
24 0  bitcoincash:q9adhakpwzztepkpwp5z0dq62m6u5v5xtyj7j3h2ws4mr9g0
24 1  bchtest:p9adhakpwzztepkpwp5z0dq62m6u5v5xtyj7j3h2u94tsynr
24 1  pref:p9adhakpwzztepkpwp5z0dq62m6u5v5xtyj7j3h2khlwwk5v
24 15 prefix:09adhakpwzztepkpwp5z0dq62m6u5v5xtyj7j3h2p29kc2lp

3A84F9CF51AAE98A3BB3A78BF16A6183790B18719126325BFC0C075B
28 0  bitcoincash:qgagf7w02x4wnz3mkwnchut2vxphjzccwxgjvvjmlsxqwkcw59jxxuz
28 1  bchtest:pgagf7w02x4wnz3mkwnchut2vxphjzccwxgjvvjmlsxqwkcvs7md7wt
28 1  pref:pgagf7w02x4wnz3mkwnchut2vxphjzccwxgjvvjmlsxqwkcrsr6gzkn
28 15 prefix:0gagf7w02x4wnz3mkwnchut2vxphjzccwxgjvvjmlsxqwkc5djw8s9g

3173EF6623C6B48FFD1A3DCC0CC6489B0A07BB47A37F47CFEF4FE69DE825C060
32 0  bitcoincash:qvch8mmxy0rtfrlarg7ucrxxfzds5pamg73h7370aa87d80gyhqxq5nlegake
32 1  bchtest:pvch8mmxy0rtfrlarg7ucrxxfzds5pamg73h7370aa87d80gyhqxq7fqng6m6
32 1  pref:pvch8mmxy0rtfrlarg7ucrxxfzds5pamg73h7370aa87d80gyhqxq4k9m7qf9
32 15 prefix:0vch8mmxy0rtfrlarg7ucrxxfzds5pamg73h7370aa87d80gyhqxqsh6jgp6w

C07138323E00FA4FC122D3B85B9628EA810B3F381706385E289B0B25631197D194B5C238BEB136FB
40 0  bitcoincash:qnq8zwpj8cq05n7pytfmskuk9r4gzzel8qtsvwz79zdskftrzxtar994cgutavfklv39gr3uvz
40 1  bchtest:pnq8zwpj8cq05n7pytfmskuk9r4gzzel8qtsvwz79zdskftrzxtar994cgutavfklvmgm6ynej
40 1  pref:pnq8zwpj8cq05n7pytfmskuk9r4gzzel8qtsvwz79zdskftrzxtar994cgutavfklv0vx5z0w3
40 15 prefix:0nq8zwpj8cq05n7pytfmskuk9r4gzzel8qtsvwz79zdskftrzxtar994cgutavfklvwsvctzqy

E361CA9A7F99107C17A622E047E3745D3E19CF804ED63C5C40C6BA763696B98241223D8CE62AD48D863F4CB18C930E4C
48 0  bitcoincash:qh3krj5607v3qlqh5c3wq3lrw3wnuxw0sp8dv0zugrrt5a3kj6ucysfz8kxwv2k53krr7n933jfsunqex2w82sl
48 1  bchtest:ph3krj5607v3qlqh5c3wq3lrw3wnuxw0sp8dv0zugrrt5a3kj6ucysfz8kxwv2k53krr7n933jfsunqnzf7mt6x
48 1  pref:ph3krj5607v3qlqh5c3wq3lrw3wnuxw0sp8dv0zugrrt5a3kj6ucysfz8kxwv2k53krr7n933jfsunqjntdfcwg
48 15 prefix:0h3krj5607v3qlqh5c3wq3lrw3wnuxw0sp8dv0zugrrt5a3kj6ucysfz8kxwv2k53krr7n933jfsunqakcssnmn

D9FA7C4C6EF56DC4FF423BAAE6D495DBFF663D034A72D1DC7D52CBFE7D1E6858F9D523AC0A7A5C34077638E4DD1A701BD017842789982041
56 0  bitcoincash:qmvl5lzvdm6km38lgga64ek5jhdl7e3aqd9895wu04fvhlnare5937w4ywkq57juxsrhvw8ym5d8qx7sz7zz0zvcypqscw8jd03f
56 1  bchtest:pmvl5lzvdm6km38lgga64ek5jhdl7e3aqd9895wu04fvhlnare5937w4ywkq57juxsrhvw8ym5d8qx7sz7zz0zvcypqs6kgdsg2g
56 1  pref:pmvl5lzvdm6km38lgga64ek5jhdl7e3aqd9895wu04fvhlnare5937w4ywkq57juxsrhvw8ym5d8qx7sz7zz0zvcypqsammyqffl
56 15 prefix:0mvl5lzvdm6km38lgga64ek5jhdl7e3aqd9895wu04fvhlnare5937w4ywkq57juxsrhvw8ym5d8qx7sz7zz0zvcypqsgjrqpnw8

D0F346310D5513D9E01E299978624BA883E6BDA8F4C60883C10F28C2967E67EC77ECC7EEEAEAFC6DA89FAD72D11AC961E164678B868AEEEC5F2C1DA08884175B
64 0  bitcoincash:qlg0x333p4238k0qrc5ej7rzfw5g8e4a4r6vvzyrcy8j3s5k0en7calvclhw46hudk5flttj6ydvjc0pv3nchp52amk97tqa5zygg96mtky5sv5w
64 1  bchtest:plg0x333p4238k0qrc5ej7rzfw5g8e4a4r6vvzyrcy8j3s5k0en7calvclhw46hudk5flttj6ydvjc0pv3nchp52amk97tqa5zygg96mc773cwez
64 1  pref:plg0x333p4238k0qrc5ej7rzfw5g8e4a4r6vvzyrcy8j3s5k0en7calvclhw46hudk5flttj6ydvjc0pv3nchp52amk97tqa5zygg96mg7pj3lh8
64 15 prefix:0lg0x333p4238k0qrc5ej7rzfw5g8e4a4r6vvzyrcy8j3s5k0en7calvclhw46hudk5flttj6ydvjc0pv3nchp52amk97tqa5zygg96ms92w6845
"""

class unit_tests:

	@property
	def vectors(self):
		t = namedtuple('vectors', ['size', 'type', 'addr', 'data'])
		def gen():
			for a in vec_data.splitlines():
				if a:
					d = a.split()
					if len(d) == 1:
						data = d[0].lower()
					else:
						yield t(int(d[0]), int(d[1]), d[2], data)
		return list(gen())

	@property
	def aliases(self):
		t = namedtuple('aliases', ['legacy', 'cashaddr'])
		return [t(*a.split()) for a in alias_data.splitlines() if a]

	def encode(self, name, ut, desc='low-level address encoding'):
		data = None
		for v in self.vectors:
			if not data or data != v.data:
				data = v.data
				vmsg(f'\n{data}')
			vmsg(f'    {v.addr}')
			ret = cashaddr_encode_addr(v.type, v.size, cashaddr_parse_addr(v.addr).pfx, bytes.fromhex(v.data))
			assert ret.addr == v.addr
		return True

	def decode(self, name, ut, desc='low-level address decoding'):
		data = None
		for v in self.vectors:
			if not data or data != v.data:
				data = v.data
				vmsg(f'\n{data}')
			vmsg(f'    {v.addr}')
			ret = cashaddr_decode_addr(v.addr)
			assert ret.bytes.hex() == v.data
		return True

	def coinaddr(self, name, ut, desc='CoinAddr class'):
		for e in self.aliases:
			for addr in (
					e.cashaddr.upper(),
					e.cashaddr,
					e.cashaddr.split(':')[1],
					e.legacy,
				):
				a = CoinAddr(proto, addr)
				vmsg(addr)
				assert e.legacy == a.views[1]
				assert e.cashaddr == a.proto.cashaddr_pfx + ':' + a.views[0]
			vmsg('')
		return True

	def errors(self, name, ut, desc='error handling'):
		# could do these in objtest.py:
		def bad1(): a = CoinAddr(proto, self.aliases[0].cashaddr.replace('g', 'G'))
		def bad2(): a = CoinAddr(proto, 'x' + self.aliases[0].cashaddr)
		def bad3(): a = CoinAddr(proto, self.aliases[0].cashaddr[:-1])
		def bad4(): a = CoinAddr(proto, self.aliases[0].cashaddr[:-1]+'i')
		def bad5(): a = CoinAddr(proto, self.aliases[0].cashaddr[:-1]+'x')

		ut.process_bad_data((
			('case',     'ObjectInitError', 'mixed case',     bad1),
			('prefix',   'ObjectInitError', 'invalid prefix', bad2),
			('data',     'ObjectInitError', 'too short',      bad3),
			('b32 char', 'ObjectInitError', 'substring',      bad4),
			('chksum',   'ObjectInitError', 'checksum',       bad5),
		))
		return True
